"""Collection of tools that can be used while testing the laserchicken software."""
import datetime as dt

import numpy as np

from laserchicken import keys


def generate_tiny_test_point_cloud():
    """Generate a simple but valid point cloud with 3 points."""
    pc = {keys.point: {'x': {'type': 'double', 'data': np.array([1, 2, 3], dtype=np.float64)},
                       'y': {'type': 'double', 'data': np.array([2, 3, 4], dtype=np.float64)},
                       'z': {'type': 'double', 'data': np.array([3, 4, 5], dtype=np.float64)}}}
    return pc


class SimpleTestData(object):
    """Test data within this class should all be in sync (reflect the same data)."""

    @staticmethod
    def get_point_cloud():
        """Get the point cloud data."""
        # This simple_test_point cloud and the simple_test_header should be in sync. Some tests depend on it.
        pc = {keys.point: {'x': {'type': 'double', 'data': np.array([1, 2, 3], dtype=np.float64)},
                           'y': {'type': 'double', 'data': np.array([20, 30, 40], dtype=np.float64)},
                           'z': {'type': 'double', 'data': np.array([300, 400, 500], dtype=np.float64)}}}
        return pc

    @staticmethod
    def get_header(is_binary=False):
        """Get the ply header."""
        if is_binary:
            format = "binary_little_endian"
        else:
            format = "ascii"
        # This simple_test_header cloud and the simple_test_point should be in sync. Some tests depend on it.
        header = """ply
format {} 1.0
element vertex 3
property double x
property double y
property double z
""".format(format)
        return header

    @staticmethod
    def get_data(is_binary=False):
        """Get the data in ply format."""
        if not is_binary:
            data = """1.0 20.0 300.0
2.0 30.0 400.0
3.0 40.0 500.0
"""
        else:
            with open("testdata/simple_test_data_little_endian.bin", "rb") as f:
                data = f.read().rstrip()
        return data


class ComplexTestData(object):
    """Test data within this class should all be in sync (reflect the same data)."""

    @staticmethod
    def get_point_cloud():
        """Get the point cloud data."""
        # This complex_test_point cloud and the complex_test_header should be in sync. Some tests depend on it.
        pc = {keys.point: {'x': {'type': 'double', 'data': np.array([1, 2, 3, 4, 5], dtype=np.float)},
                           'y': {'type': 'double', 'data': np.array([2, 3, 4, 5, 6], dtype=np.float)},
                           'z': {'type': 'double', 'data': np.array([3, 4, 5, 6, 7], dtype=np.float)},
                           'return': {'type': 'int', 'data': np.array([1, 1, 2, 2, 1], dtype=np.int32)}
                           },
              keys.point_cloud: {'offset': {'type': 'double', 'data': 12.1}},
              keys.provenance: [{"module": "filter", "time": str(dt.datetime(2018, 1, 18, 16, 1, 0))},
                                {"module": "filter", "time": str(dt.datetime(2018, 2, 4, 13, 11, 0))}]
              }
        return pc

    @staticmethod
    def get_header(is_binary=False):
        """Get the ply header."""
        if is_binary:
            format = "binary_little_endian"
        else:
            format = "ascii"
        # This complex_test_header cloud and the complex_test_point should be in sync. Some tests depend on it.
        header = ("""ply
format {} 1.0
comment [
comment {{"module": "filter", "time": "2018-01-18 16:01:00"}},
comment {{"module": "filter", "time": "2018-02-04 13:11:00"}}
comment ]
element vertex 5
property double x
property double y
property double z
property int return
element pointcloud 1
property double offset
""").format(format)
        return header

    @staticmethod
    def get_data(is_binary=False):
        """Get the data in ply format."""
        if not is_binary:
            data = """1.0 2.0 3.0 1
2.0 3.0 4.0 1
3.0 4.0 5.0 2
4.0 5.0 6.0 2
5.0 6.0 7.0 1
12.1
"""
        else:
            with open("testdata/complex_test_data_little_endian.bin", "rb") as f:
                data = f.read().rstrip()
        return data

    @staticmethod
    def get_wkt_polygon_around_first_point_only():
        """Get a polygon in wkt format that surrounds only the first point."""
        return "POLYGON(( 1.5 10.0, 1.5 -10.0, -1.5 -10.0, -1.5 1.0, 1.5 10.0 ))"


def create_point_cloud(x, y, z, normalized_z=None):
    """
    Create a point cloud object given only the x y z values.

    :param x: x attribute values
    :param y: y attribute values
    :param z: z attribute values
    :param normalized_z: optional normalized z attribute values
    :return: point cloud object
    """
    point_cloud = {keys.point: {'x': {'type': 'double', 'data': np.array(x, dtype=np.float)},
                                'y': {'type': 'double', 'data': np.array(y, dtype=np.float)},
                                'z': {'type': 'double', 'data': np.array(z, dtype=np.float)}},
                   keys.point_cloud: {},
                   keys.provenance: [{'time': (dt.datetime(2018, 1, 18, 16, 1, 0)), 'module': 'filter'}]}
    if normalized_z is not None:
        point_cloud[keys.point][keys.normalized_height] = {'type': 'double',
                                                           'data': np.array(normalized_z, dtype=np.float)}
    return point_cloud


def create_points_in_xy_grid(z_function):
    """Create 100 points in a grid of 10 by 10."""
    n_points = 100
    points = np.zeros((n_points, 3))
    for i in range(n_points):
        x = i % np.sqrt(n_points)
        y = np.floor(i / np.sqrt(n_points))
        points[i] = np.array((x, y, z_function(x, y)))
    return n_points, points



import os
import unittest
import random

import pytest
import numpy as np

from laserchicken import keys, load, utils
from laserchicken.compute_neighbors import compute_neighborhoods
from laserchicken.feature_extractor.base_feature_extractor import FeatureExtractor
from laserchicken.keys import point
from laserchicken.test_tools import create_point_cloud
from laserchicken.volume_specification import Sphere, InfiniteCylinder

from laserchicken.feature_extractor.echo_ratio_feature_extractor import EchoRatioFeatureExtractor


class TestEchoRatioFeatureExtractorArtificialData(unittest.TestCase):
    """Test echo ratio extractor on artificial spherical and cylindric data."""

    point_cloud = None
    target_point_cloud = None

    def test_valid(self):
        """Result must be close to the theoretical value."""

        # extractor = EchoRatioFeatureExtractor()
        extractor = EchoRatioFeatureExtractor()
        per = extractor.extract(
            self.point_cloud, self.neighborhoods, self.target_point_cloud, self.indexpc, self.cylinder)
        np.testing.assert_allclose(per, self.theoretical_value)

    def test_invalid(self):
        """ Must raise TypeError as we do not provide correct indexes."""

        extractor = EchoRatioFeatureExtractor()
        # target point cloud must not be None
        with pytest.raises(ValueError):
            extractor.extract(self.point_cloud, self.neighborhoods, None, self.indexpc, self.cylinder)

        # target index must not be None
        with pytest.raises(ValueError):
            extractor.extract(self.point_cloud, self.neighborhoods, self.target_point_cloud, None, self.cylinder)

        # volume must be a cylinder
        with pytest.raises(ValueError):
            sphere = Sphere(self.radius)
            extractor.extract(self.point_cloud, self.neighborhoods, self.target_point_cloud, self.indexpc, sphere)

    @staticmethod
    def _get_pc(xyz):
        return {keys.point: {'x': {'type': 'double', 'data': xyz[:, 0]},
                             'y': {'type': 'double', 'data': xyz[:, 1]},
                             'z': {'type': 'double', 'data': xyz[:, 2]}}}, len(xyz)

    def _set_sphere_data(self):
        """Create a sphere of point."""
        n_theta, n_phi = 11, 11
        self.npt_sphere = n_theta * n_phi
        theta = np.linspace(0.1, 2 * np.pi, n_theta)
        phi = np.linspace(0.1, np.pi, n_phi)
        r = self.radius
        for t in theta:
            for p in phi:
                x = r * np.cos(t) * np.sin(p)
                y = r * np.sin(t) * np.sin(p)
                z = r * np.cos(p)
                self.xyz.append([x, y, z])

    def _set_cylinder_data(self):
        """Create a cylinder of point."""
        # n_height must be even so that the cylinder has no point
        # on the equator of the sphere
        n_theta, n_height = 11, 10
        self.npt_cyl = n_theta * n_height
        r = self.radius
        theta = np.linspace(0.1, 2 * np.pi, n_theta)
        height = np.linspace(-2 * r, 2 * r, n_height)

        for h in height:
            for t in theta:
                x, y, z = r * np.cos(t), r * np.sin(t), h
                self.xyz.append([x, y, z])

    def _get_central_point(self, index):
        """Get the central point."""
        return utils.copy_point_cloud(self.point_cloud, [index])

    def setUp(self):
        """
        Set up the test.

        Create a sphere and a cylinder of points and a central point
        The cylinder has no points on the equator of the sphere
        """

        # create the points
        self.radius = 0.5
        self.xyz = [[0., 0., 0.]]
        self._set_sphere_data()
        self._set_cylinder_data()

        # create the pc
        self.point_cloud, self.npts = self._get_pc(np.array(self.xyz))
        self.target_point_cloud = self._get_central_point(0)
        self.indexpc = 0

        # create the volume/neighborhood
        self.cylinder = InfiniteCylinder(self.radius + 1E-3)
        self.neighborhoods = list(compute_neighborhoods(self.point_cloud, self.target_point_cloud, self.cylinder))

        # theoretical value of the echo ratio
        self.theoretical_value = (self.npt_sphere + 1) / (self.npt_sphere + self.npt_cyl + 1)


class TestEchoRatioFeatureExtractorSimpleArtificialData(unittest.TestCase):
    """Test echo ratio extractor on artificial spherical and cylindric data."""

    def test_valid(self):
        """Result must be close to the theoretical value."""
        extractor = EchoRatioFeatureExtractor()
        result = extractor.extract(self.environment_pc, self.neighbors, self.target_pc, range(4), self.cylinder)
        np.testing.assert_allclose(result, self.echo_ratios)

    def setUp(self):
        """
        Create a grid of 4 targets and an environment point cloud and neighbors and the echo ratios of those targets.
        """
        self.radius = 0.5
        targets = np.array([[10., 0., 5.], [10., 10., 5.], [0., 0., 5.], [0., 10., 5.]])  # Grid w. steps 10 & height 5
        self.echo_ratios = np.array([0., 0.9, 0.5, 0.8])
        environment_parts = [self._create_environment_part(t, ratio, self.radius) for t, ratio in zip(targets,
                                                                                                      self.echo_ratios)]
        environment = np.vstack(environment_parts)

        self.target_pc = create_point_cloud(targets[:, 0], targets[:, 1], targets[:, 2])
        self.environment_pc = create_point_cloud(environment[:, 0], environment[:, 1], environment[:, 2])

        self.cylinder = InfiniteCylinder(self.radius)
        self.neighbors = list(compute_neighborhoods(self.environment_pc, self.target_pc, self.cylinder))

    @staticmethod
    def _create_environment_part(target, echo_ratio, radius):
        """
        Creates 10 points at the target location, and a given ratio of points outside the radius (above the target)
        :param target:
        :param echo_ratio:
        :param radius:
        :return: numpy array
        """
        ratio = echo_ratio  # Convert from percentage
        n_outside = 10
        n_inside = int(n_outside * ratio / (1 - ratio))
        outside_sphere = np.zeros((n_outside, 3)) + target + np.array([0., 0., 2 * radius])
        inside_sphere = np.zeros((int(n_inside), 3)) + target
        return np.vstack((inside_sphere, outside_sphere))


class EchoRatioFeatureExtractorSequential(FeatureExtractor):
    """Feature extractor for the point density."""

    @classmethod
    def requires(cls):
        """
        Get a list of names of the point attributes that are needed for this feature extraction.
        For simple features, this could be just x, y, and z. Other features can build on again
        other features to have been computed first.
        :return: List of feature names
        """
        return []

    @classmethod
    def provides(cls):
        """
        Get a list of names of the feature values.
        This will return as many names as the number feature values that will be returned.
        For instance, if a feature extractor returns the first 3 Eigen values, this method
        should return 3 names, for instance 'eigen_value_1', 'eigen_value_2' and 'eigen_value_3'.
        :return: List of feature names
        """
        return ['echo_ratio_sequential']

    def extract(self, point_cloud, neighborhood, target_point_cloud, target_index, volume_description):
        """
        Extract the feature value(s) of the point cloud at location of the target.
        :param point_cloud: environment (search space) point cloud
        :param neighborhood: array of indices of points within the point_cloud argument
        :param target_point_cloud: point cloud that contains target point
        :param target_index: index of the target point in the target point cloud
        :param volume_description: volume object that describes the shape and size of the search volume
        :return: feature value
        """
        if volume_description.TYPE != 'infinite cylinder':
            raise ValueError('The volume must be a cylinder')

        if target_point_cloud is None:
            raise ValueError('Target point cloud required')

        if target_index is None:
            raise ValueError('Target point index required')

        xyz = self.get_neighborhood_positions(point_cloud, neighborhood)
        n_cylinder = xyz.shape[0]

        xyz0 = self.get_target_position(target_point_cloud, target_index)
        difference = xyz - xyz0
        squared = difference ** 2
        sum_of_squares = np.sum(squared, 1)
        n_sphere = np.sum(sum_of_squares <=
                          volume_description.radius ** 2)
        return n_sphere / n_cylinder

    @staticmethod
    def get_target_position(target_point_cloud, target_index):
        x0 = target_point_cloud[point]['x']['data'][target_index]
        y0 = target_point_cloud[point]['y']['data'][target_index]
        z0 = target_point_cloud[point]['z']['data'][target_index]
        return np.array([x0, y0, z0])

    @staticmethod
    def get_neighborhood_positions(point_cloud, neighborhood):
        x = point_cloud[point]['x']['data'][neighborhood]
        y = point_cloud[point]['y']['data'][neighborhood]
        z = point_cloud[point]['z']['data'][neighborhood]
        return np.column_stack((x, y, z))

    def get_params(self):
        """
        Return a tuple of parameters involved in the current feature extractorobject.
        Needed for provenance.
        """
        return ()


if __name__ == '__main__':
    unittest.main()



"""Sigma Z feature extractor.

Here, Sigma Z is defined as the standard deviation of the residuals after plane fitting.
See https://github.com/eEcoLiDAR/eEcoLiDAR/issues/20
"""
import numpy as np
from numpy.linalg import LinAlgError

from laserchicken.feature_extractor.base_feature_extractor import FeatureExtractor
from laserchicken.utils import get_point, fit_plane


class SigmaZFeatureExtractor(FeatureExtractor):
    """Height percentiles feature extractor class."""

    @classmethod
    def requires(cls):
        """
        Get a list of names of the point attributes that are needed for this feature extraction.

        For simple features, this could be just x, y, and z. Other features can build on again
        other features to have been computed first.

        :return: List of feature names
        """
        return []

    @classmethod
    def provides(cls):
        """
        Get a list of names of the feature values.

        This will return as many names as the number feature values that will be returned.
        For instance, if a feature extractor returns the first 3 Eigen values, this method
        should return 3 names, for instance 'eigen_value_1', 'eigen_value_2' and 'eigen_value_3'.

        :return: List of feature names
        """
        return ['sigma_z']

    def extract(self, point_cloud, neighborhoods, target_point_cloud, target_indices, volume_description):
        """
        Extract the feature value(s) of the point cloud at location of the target.

        :param point_cloud: environment (search space) point cloud
        :param neighborhoods: list of arrays of indices of points within the point_cloud argument
        :param target_point_cloud: point cloud that contains target point
        :param target_indices: list of indices of the target point in the target point cloud
        :param volume_description: volume object that describes the shape and size of the search volume
        :return: feature values
        """
        return [self._extract_one(point_cloud, neighborhood) for neighborhood in neighborhoods]

    @staticmethod
    def _extract_one(source_point_cloud, neighborhood):
        """
        Extract the feature value(s) of the point cloud at location of the target.

        :param source_point_cloud: environment (search space) point cloud
        :param neighborhood: array of indices of points within the point_cloud argument
        :return:
        """
        x, y, z = get_point(source_point_cloud, neighborhood)
        try:
            plane_estimator = fit_plane(x, y, z)
            normalized = z - plane_estimator(x, y)
            return np.std(normalized)
        except LinAlgError:
            return 0

    def get_params(self):
        """
        Return a tuple of parameters involved in the current feature extractor object.

        Needed for provenance.
        """
        return ()



"""Shannan entropy calculation. For more info see https://rdrr.io/cran/lidR/man/entropy.html"""

import numpy as np
from laserchicken import keys
from laserchicken.feature_extractor.base_feature_extractor import FeatureExtractor


class EntropyFeatureExtractor(FeatureExtractor):
    layer_thickness = 0.5
    min_val = None
    max_val = None

    def __init__(self, data_key='z'):
        self.data_key = data_key

    @classmethod
    def requires(cls):
        return []

    def provides(self):
        return ['entropy_' + self.data_key]

    def get_params(self):
        return [self.layer_thickness, self.min_val, self.max_val]

    def extract(self, point_cloud, neighborhoods, target_point_cloud, target_indices, volume):
        return [self._extract_one(point_cloud, neighborhood) for neighborhood in neighborhoods]

    def _extract_one(self, source_pc, neighborhood):
        if len(neighborhood) == 0:
            return 0
        source_data = source_pc[keys.point][self.data_key]["data"][neighborhood]
        data_min = np.min(source_data) if self.min_val is None else self.min_val
        data_max = np.max(source_data) if self.max_val is None else self.max_val
        if data_min == data_max:
            return 0
        n_bins = int(np.ceil((data_max - data_min) / self.layer_thickness))
        data = np.histogram(source_data, bins=n_bins, range=(data_min, data_max), density=True)[0]
        entropy_func = np.vectorize(_x_log_2x)
        norm = np.sum(data)
        return -(entropy_func(data / norm)).sum()


def _x_log_2x(x):
    return 0 if x == 0 else x * np.log2(x)



"""Calculate echo ratio.

See https://github.com/eEcoLiDAR/eEcoLiDAR/issues/21
"""

import numpy as np

from laserchicken.feature_extractor.base_feature_extractor import FeatureExtractor
from laserchicken.utils import get_attributes_per_neighborhood


def _to_unmasked_array(masked_array):
    """Creates a 'normal' numpy array from a masked array, inputting nans for masked values."""
    data = masked_array.data
    data[masked_array.mask] = np.nan
    return data


class BandRatioFeatureExtractor(FeatureExtractor):
    """Feature extractor for the point density."""

    def __init__(self, lower_limit, upper_limit, data_key='z'):
        self.lower_limit = lower_limit
        self.upper_limit = upper_limit
        self.data_key = data_key

    def requires(self):
        """
        Get a list of names of the point attributes that are needed for this feature extraction.

        For simple features, this could be just x, y, and z. Other features can build on again
        other features to have been computed first.

        :return: List of feature names
        """
        return []

    def provides(self):
        """
        Get a list of names of the feature values.

        This will return as many names as the number feature values that will be returned.
        For instance, if a feature extractor returns the first 3 Eigen values, this method
        should return 3 names, for instance 'eigen_value_1', 'eigen_value_2' and 'eigen_value_3'.

        :return: List of feature names
        """
        name = 'band_ratio_'
        if self.lower_limit is not None:
            name += str(self.lower_limit) + '<'
        name += self.data_key
        if self.upper_limit is not None:
            name += '<' + str(self.upper_limit)
        return [name]

    def extract(self, point_cloud, neighborhoods, target_point_cloud, target_index, volume_description):
        """
        Extract the feature value(s) of the point cloud at location of the target.

        :param point_cloud: environment (search space) point cloud
        :param neighborhoods: array of array of indices of points within the point_cloud argument
        :param target_point_cloud: point cloud that contains target point
        :param target_index: index of the target point in the target point cloud
        :param volume_description: volume object that describes the shape and size of the search volume
        :return: feature value
        """
        supported_volumes = ['infinite cylinder', 'cell']
        if volume_description.TYPE not in supported_volumes:
            raise ValueError('The volume must be a cylinder')

        attribute = get_attributes_per_neighborhood(point_cloud, neighborhoods, [self.data_key])
        z = attribute[:, 0, :]
        n_total_points = attribute.shape[2]
        n_masked_points_per_neighborhood = attribute.mask[:, 0, :].sum(axis=1)
        n_points_per_neighborhood = -n_masked_points_per_neighborhood + n_total_points
        is_point_below_upper_limit = z < self.upper_limit if self.upper_limit else np.ones_like(z)
        is_point_above_lower_limit = z > self.lower_limit if self.lower_limit else np.ones_like(z)
        n_points_within_band = np.sum(is_point_below_upper_limit * is_point_above_lower_limit, axis=1)
        return _to_unmasked_array(n_points_within_band / n_points_per_neighborhood)

    def get_params(self):
        """
        Return a tuple of parameters involved in the current feature extractor object.

        Needed for provenance.
        """
        return (self.lower_limit, self.upper_limit, self.data_key)



import os
import unittest

import random
import numpy as np

from laserchicken import keys, load, utils
from laserchicken.compute_neighbors import compute_neighborhoods
from laserchicken.volume_specification import Sphere, InfiniteCylinder, Cell, Cube
from laserchicken.feature_extractor.density_feature_extractor import PointDensityFeatureExtractor
from laserchicken.test_tools import create_point_cloud


class TestDensityFeatureExtractorSphere(unittest.TestCase):
    """Test density extractor on artificial spherical data."""

    point_cloud = None

    def test_sphere(self):
        """Compute the density for a sphere given as index of the source pc."""
        neighborhoods = list(compute_neighborhoods(self.point_cloud, self.targetpc, self.sphere))

        extractor = PointDensityFeatureExtractor()
        densities = extractor.extract(self.point_cloud, neighborhoods, None, None, self.sphere)
        np.testing.assert_allclose(densities, self.volume_density)

    def _get_central_point(self):
        """Get the central point."""
        return utils.copy_point_cloud(self.point_cloud, [0])

    def _set_sphere_data(self):
        """Generate a pc of points regularly positionned on a two spheres of radius 1 and 2."""
        nteta, nphi = 11, 11
        self.teta = np.linspace(0.1, 2 * np.pi, nteta)
        self.phi = np.linspace(0.1, np.pi, nphi)
        self.radius = [1., 2.]
        self.points_per_sphere = nteta * nphi
        self.xyz = []
        self.xyz.append([0., 0., 0.])
        for r in self.radius:
            for t in self.teta:
                for p in self.phi:
                    x = r * np.cos(t) * np.sin(p)
                    y = r * np.sin(t) * np.sin(p)
                    z = r * np.cos(p)
                    self.xyz.append([x, y, z])

        self.xyz = np.array(self.xyz)
        self.point_cloud = {keys.point: {'x': {'type': 'double', 'data': self.xyz[:, 0]},
                                         'y': {'type': 'double', 'data': self.xyz[:, 1]},
                                         'z': {'type': 'double', 'data': self.xyz[:, 2]}}}

    def setUp(self):
        """Set up the test."""
        # get the points
        self._set_sphere_data()

        # get the central point as targetpc
        self.targetpc = self._get_central_point()

        # get the sphere
        self.sphere = Sphere(np.mean(self.radius))

        # get the theoretical value +1 for central point
        npts = self.points_per_sphere + 1
        self.volume_density = npts / self.sphere.calculate_volume()

    def tearDown(self):
        pass


class TestDensityFeatureExtractorCylinder(unittest.TestCase):
    """Test density extractor on artificial cylindric data."""

    point_cloud = None

    def test_cylinder(self):
        """Compute the density for a cylinder given as index of source pc."""
        neighborhoods = compute_neighborhoods(self.point_cloud, self.targetpc, self.cyl)

        extractor = PointDensityFeatureExtractor()
        densities = extractor.extract(self.point_cloud, neighborhoods, None, None, self.cyl)
        np.testing.assert_allclose(densities, self.real_density)

    def _get_central_point(self):
        """Get the central point."""
        return utils.copy_point_cloud(self.point_cloud, [0])

    def _set_cylinder_data(self):

        # generate a pc of points regularly
        # positionned on two concentric cylinders
        # plus a central point
        nteta, nheight = 11, 11
        teta = np.linspace(0.01, 2 * np.pi, nteta)
        height = np.linspace(-1, 1, nheight)
        self.radius = [1., 2.]

        self.points_per_cylinder = nteta * nheight
        self.xyz = []
        self.xyz.append([0., 0., 0.])
        for r in self.radius:
            for z in height:
                for t in teta:
                    x = r * np.cos(t)
                    y = r * np.sin(t)
                    self.xyz.append([x, y, z])

        self.xyz = np.array(self.xyz)
        self.point_cloud = {keys.point: {'x': {'type': 'double', 'data': self.xyz[:, 0]},
                                         'y': {'type': 'double', 'data': self.xyz[:, 1]},
                                         'z': {'type': 'double', 'data': self.xyz[:, 2]}}}

    def setUp(self):
        """Set up the test."""
        # generate the points
        self._set_cylinder_data()

        # get the central point as target pc
        self.targetpc = self._get_central_point()

        # get the cylinder
        self.cyl = InfiniteCylinder(np.mean(self.radius))

        # get the theoretical value +1 for central point
        n_points = self.points_per_cylinder + 1
        self.real_density = n_points / self.cyl.calculate_base_area()

    def tearDowm(self):
        pass


class TestDensityFeatureForCell(unittest.TestCase):
    def test_cell(self):
        n_included = 123
        n_excluded = 456
        x = np.append(np.zeros(n_included), np.ones(n_excluded))
        environment = create_point_cloud(x, x, x)
        target = create_point_cloud(np.zeros(1), np.zeros(1), np.zeros(1))

        cell = Cell(1)  # area = 1.0
        neighborhoods = compute_neighborhoods(environment, target, cell)

        extractor = PointDensityFeatureExtractor()
        densities = extractor.extract(environment, neighborhoods, target, [0], cell)
        np.testing.assert_allclose(densities, n_included)


class TestDensityFeatureForCube(unittest.TestCase):
    def test_cell(self):
        n_included = 123
        n_excluded = 456
        x = np.append(np.zeros(n_included), np.ones(n_excluded))
        environment = create_point_cloud(x, x, x)
        target = create_point_cloud(np.zeros(1), np.zeros(1), np.zeros(1))

        cube = Cube(1)  # volume = 1.0

        neighborhoods = compute_neighborhoods(environment, target, cube)
        extractor = PointDensityFeatureExtractor()
        densities = extractor.extract(environment, neighborhoods, target, [0], cube)
        np.testing.assert_allclose(densities, n_included)


class TestDensityFeatureOnRealData(unittest.TestCase):
    """Test density extractor on real data and make sure it doesn't crash."""

    _test_file_name = 'AHN3.las'
    _test_data_source = 'testdata'
    point_cloud = None

    def test_sphere_index(self):
        """Compute the density for a sphere given as index of the source pc."""
        neighborhoods = compute_neighborhoods(self.point_cloud, self.target_point_cloud, self.sphere)
        extractor = PointDensityFeatureExtractor()
        extractor.extract(self.point_cloud, neighborhoods, None, None, self.sphere)

    def test_cylinder_index(self):
        """Compute the density for a cylinder given as index of source pc."""
        neighborhoods = compute_neighborhoods(self.point_cloud, self.target_point_cloud, self.cyl)
        extractor = PointDensityFeatureExtractor()
        extractor.extract(self.point_cloud, neighborhoods, None, None, self.cyl)

    def _get_random_targets(self):
        num_all_pc_points = len(self.point_cloud[keys.point]["x"]["data"])
        rand_indices = [random.randint(0, num_all_pc_points)
                        for p in range(20)]
        return utils.copy_point_cloud(self.point_cloud, rand_indices)

    def setUp(self):
        """Set up the test."""
        self.point_cloud = load(os.path.join(self._test_data_source, self._test_file_name))

        random.seed(102938482634)
        self.target_point_cloud = self._get_random_targets()

        radius = 0.5
        self.sphere = Sphere(radius)
        self.cyl = InfiniteCylinder(radius)

    def tearDown(self):
        pass


if __name__ == '__main__':
    unittest.main()



import numpy as np
import unittest
import datetime
import pytest
from laserchicken import utils, test_tools, keys
from laserchicken.utils import fit_plane
from time import time

class TestUtils(unittest.TestCase):
    def test_GetPointCloudPoint(self):
        """ Should not raise exception. """
        pc = test_tools.generate_tiny_test_point_cloud()
        x, y, z = utils.get_point(pc, 1)
        self.assertEqual(2, x)
        self.assertEqual(3, y)
        self.assertEqual(4, z)

    def test_GetPointCloudPointFeature(self):
        """ Should not raise exception. """
        pc = test_tools.generate_tiny_test_point_cloud()
        cols = 0.5 * (pc[keys.point]["x"]["data"] + pc[keys.point]["y"]["data"])
        pc[keys.point]["color"] = {"type": "double", "data": cols}
        x, y, z = utils.get_point(pc, 1)
        c = utils.get_attribute_value(pc, 1, "color")
        self.assertEqual(c, 0.5 * (x + y))

    def test_GetPointCloudPointFeatures(self):
        """ Should not raise exception. """
        pc = test_tools.generate_tiny_test_point_cloud()
        cols = 0.5 * (pc[keys.point]["x"]["data"] + pc[keys.point]["y"]["data"])
        flavs = 0.5 * (pc[keys.point]["x"]["data"] - pc[keys.point]["y"]["data"])
        pc[keys.point]["color"] = {"type": "double", "data": cols}
        pc[keys.point]["flavor"] = {"type": "double", "data": flavs}
        x, y, z = utils.get_point(pc, 2)
        c, f = utils.get_features(pc, ("color", "flavor"), 2)
        self.assertEqual(c, 0.5 * (x + y))
        self.assertEqual(f, 0.5 * (x - y))

    def test_CopyEmptyPointCloud(self):
        """ Should not raise exception. """
        pc = test_tools.generate_tiny_test_point_cloud()
        pc[keys.point]["x"]["data"] = np.array([])
        pc[keys.point]["y"]["data"] = np.array([])
        pc[keys.point]["z"]["data"] = np.array([])

        copypc = utils.copy_point_cloud(pc)
        self.assertEqual(0, len(copypc[keys.point]["x"]["data"]))

    def test_CopyNonEmptyPointCloud(self):
        """ Test whether coordinates are copied """
        pc = test_tools.generate_tiny_test_point_cloud()
        x = pc[keys.point]["x"]["data"]
        y = pc[keys.point]["y"]["data"]
        z = pc[keys.point]["z"]["data"]

        copypc = utils.copy_point_cloud(pc)
        self.assertTrue(all(x == copypc[keys.point]["x"]["data"]))
        self.assertTrue(all(y == copypc[keys.point]["y"]["data"]))
        self.assertTrue(all(z == copypc[keys.point]["z"]["data"]))

    def test_CopyPointCloudMetaData(self):
        """ Test whether metadata are copied """
        pc = test_tools.generate_tiny_test_point_cloud()
        pc["log"] = [
            {"time": datetime.datetime(2018, 1, 23, 12, 15, 59), "module": "filter", "parameters": [("z", "gt", 0.5)]}]

        copypc = utils.copy_point_cloud(pc)
        self.assertEqual(datetime.datetime(2018, 1, 23, 12, 15, 59), copypc["log"][0]["time"])
        self.assertEqual("filter", copypc["log"][0]["module"])
        self.assertEqual([("z", "gt", 0.5)], copypc["log"][0]["parameters"])

    def test_CopyNonEmptyPointCloudBoolMask(self):
        """ Test whether coordinates are copied with boolean mask """
        pc = test_tools.generate_tiny_test_point_cloud()
        x = pc[keys.point]["x"]["data"][2]
        y = pc[keys.point]["y"]["data"][2]
        z = pc[keys.point]["z"]["data"][2]

        copypc = utils.copy_point_cloud(pc, array_mask=np.array([False, False, True]))
        self.assertTrue(all(np.array([x]) == copypc[keys.point]["x"]["data"]))
        self.assertTrue(all(np.array([y]) == copypc[keys.point]["y"]["data"]))
        self.assertTrue(all(np.array([z]) == copypc[keys.point]["z"]["data"]))

    def test_CopyNonEmptyPointCloudIntMask(self):
        """ Test whether coordinates are copied with array indexing """
        pc = test_tools.generate_tiny_test_point_cloud()
        x0, x1 = pc[keys.point]["x"]["data"][0], pc[keys.point]["x"]["data"][1]
        y0, y1 = pc[keys.point]["y"]["data"][0], pc[keys.point]["y"]["data"][1]
        z0, z1 = pc[keys.point]["z"]["data"][0], pc[keys.point]["z"]["data"][1]

        copypc = utils.copy_point_cloud(pc, array_mask=np.array([1, 0]))
        self.assertTrue(all(np.array([x1, x0]) == copypc[keys.point]["x"]["data"]))
        self.assertTrue(all(np.array([y1, y0]) == copypc[keys.point]["y"]["data"]))
        self.assertTrue(all(np.array([z1, z0]) == copypc[keys.point]["z"]["data"]))

    def test_AddMetaDataToPointCloud(self):
        """ Test adding info to the point cloud for test module """
        pc = test_tools.generate_tiny_test_point_cloud()
        from laserchicken import filter as somemodule
        utils.add_metadata(pc,somemodule,params = (0.5,"cylinder",4))
        self.assertEqual(len(pc[keys.provenance]),1)

    def test_AddToPointCloudEmpty(self):
        pc_1 = utils.create_point_cloud([],[],[])
        pc_2 = test_tools.generate_tiny_test_point_cloud()
        utils.add_to_point_cloud(pc_1, pc_2)
        for attr in pc_2[keys.point].keys():
            self.assertIn(attr, pc_1[keys.point])
            self.assertEqual(pc_1[keys.point][attr]['type'],
                             pc_2[keys.point][attr]['type'])
            self.assertTrue(all(pc_1[keys.point][attr]['data'] == pc_2[keys.point][attr]['data']))

    def test_AddToPointCloudInvalid(self):
        pc_1 = test_tools.SimpleTestData.get_point_cloud()
        # invalid format
        pc_2 = {}
        with pytest.raises(TypeError):
            utils.add_to_point_cloud(pc_1, pc_2)
        with pytest.raises(AttributeError):
            utils.add_to_point_cloud(pc_2, pc_1)
        # non-matching attributes
        test_data = test_tools.ComplexTestData()
        pc_2 = test_data.get_point_cloud()
        with pytest.raises(AttributeError):
            utils.add_to_point_cloud(pc_1, pc_2)
        # different structure
        pc_2 = {'vertex':{'x':1, 'y':2, 'z':3}}
        with pytest.raises(TypeError):
            utils.add_to_point_cloud(pc_1, pc_2)
        # different data types
        pc_2 = {'vertex': {'x': {'data': np.zeros(3, dtype=int), 'type': 'int'},
                           'y': {'data': np.zeros(3, dtype=int), 'type': 'int'},
                           'z': {'data': np.zeros(3, dtype=int), 'type': 'int'}}}
        with pytest.raises(ValueError):
            utils.add_to_point_cloud(pc_1, pc_2)

    def test_AddToPointCloud(self):
        test_data = test_tools.ComplexTestData()
        pc_source = test_data.get_point_cloud()
        pc_dest = utils.copy_point_cloud(pc_source)
        utils.add_to_point_cloud(pc_dest, pc_source)
        for key in pc_source.keys():
            self.assertIn(key, pc_dest)
        for attr in pc_source[keys.point].keys():
            self.assertEqual(len(pc_dest[keys.point][attr]['data']),
                             2*len(pc_source[keys.point][attr]['data']))
        self.assertEqual(pc_dest[keys.provenance][-1]['module'],
                         'laserchicken.utils')
    
    def test_AddFeatureArray(self):
        test_data = test_tools.ComplexTestData()
        pc = test_data.get_point_cloud()
        feature_add = np.array([1, 1, 1, 1, 1], dtype=int)
        utils.update_feature(pc, 'test_feature', feature_add)
        self.assertIn('test_feature', pc[keys.point])
        self.assertTrue(all(pc[keys.point]['test_feature']['data'] == feature_add))

    def test_AddFeatureArrayInvalid(self):
        test_data = test_tools.ComplexTestData()
        pc = test_data.get_point_cloud()
        feature_add = np.array([1, 1, 1, 1, 1, 2], dtype=int)
        with pytest.raises(AssertionError):
            utils.update_feature(pc, 'test_feature', feature_add)
    
    def test_AddFeatureArrayMask(self):
        test_data = test_tools.ComplexTestData()
        pc = test_data.get_point_cloud()
        feature_add = np.array([1, 2, 3, 4], dtype=int)
        mask = np.array([1, 1, 0, 1, 1], dtype=bool)
        utils.update_feature(pc, 'test_feature', feature_add, array_mask=mask)
        self.assertIn('test_feature', pc[keys.point])
        self.assertTrue(all(pc[keys.point]['test_feature']['data'] == [1, 2, 0, 3, 4]))
    
    def test_AddFeatureArrayMaskInvalid(self):
        test_data = test_tools.ComplexTestData()
        pc = test_data.get_point_cloud()
        feature_add = np.array([1, 2, 3, 4], dtype=int)
        mask = np.array([1, 1, 1, 1, 1], dtype=bool)
        with pytest.raises(AssertionError):
            utils.update_feature(pc, 'test_feature', feature_add, array_mask=mask)

    def test_AddFeatureValueMask(self):
        test_data = test_tools.ComplexTestData()
        pc = test_data.get_point_cloud()
        feature_add = 1.1
        mask = np.array([1, 1, 0, 1, 1], dtype=bool)
        utils.update_feature(pc, 'test_feature', feature_add, array_mask=mask)
        self.assertIn('test_feature', pc[keys.point])
        self.assertTrue(all(pc[keys.point]['test_feature']['data'] == [1.1, 1.1, 0.0, 1.1, 1.1]))
    
    def test_AddFeatureValueMaskInvalid(self):
        test_data = test_tools.ComplexTestData()
        pc = test_data.get_point_cloud()
        feature_add = 1.1
        mask = np.array([1, 1, 0, 1, 1, 1], dtype=bool)
        with pytest.raises(AssertionError):
            utils.update_feature(pc, 'test_feature', feature_add, array_mask=mask)
        

        

class TestPlaneFit(unittest.TestCase):

    def test_leastsqr(self):
        # n_points = 100
        # points = np.zeros((n_points, 3))
        # for i in range(n_points):
        #     z = 5 + i % np.sqrt(n_points)
        #     points[i] = np.array(((i % np.sqrt(n_points)), (np.floor(i / np.sqrt(n_points))), z))
        #t0 = time()
        f = fit_plane(self.points[:, 0], self.points[:, 1], self.points[:, 2])
        #print('LSQR : %f' %(time()-t0))
        estimates = f(self.points[:, 0], self.points[:, 1])
        np.testing.assert_allclose(estimates, self.points[:, 2])


    def test_FitPlaneSVD(self):
        """Test the normal vector extraction with SVD."""
        #t0 = time()
        nfit = utils.fit_plane_svd(self.points[:,0], self.points[:,1], self.points[:,2])
        #print('SVD  : %f' %(time()-t0))
        self.assertTrue(np.allclose(nfit, self.n))


    def generate_random_points_inplane(self,nvect, dparam=0, npts=100, eps=0.0):
        """
        Generate a series of point all belonging to a plane.

        :param nvect: normal vector of the plane
        :param dparam: zero point value of the plane
        :param npts: number of points
        :param eps: std of the gaussian noise added to the z values of the planes
        :return: x,y,z coordinate of the points
        """
        a, b, c = nvect / np.linalg.norm(nvect)
        x, y = np.random.rand(npts), np.random.rand(npts)
        z = (dparam - a * x - b * y) / c
        if eps > 0:
            z += np.random.normal(loc=0., scale=eps, size=npts)
        return np.column_stack((x, y, z))


    def setUp(self):
        """Set up the data points."""

        self.n = np.array([1., 2., 3.])
        self.n /= np.linalg.norm(self.n)
        self.points = self.generate_random_points_inplane(self.n,eps=0)

    def tearDown(self):
        """Tear it down."""
        pass

if __name__ == '__main__':
    unittest.main()



import json
import os

import numpy as np
import pytest

from laserchicken import compute_features, compute_neighborhoods, keys, load
from laserchicken.keys import point
#from laserchicken.feature_extractor import *
from laserchicken.feature_extractor.pulse_penetration_feature_extractor import GROUND_TAGS
from laserchicken.utils import copy_point_cloud
from laserchicken.volume_specification import InfiniteCylinder
from .feature_map import create_default_feature_map, _create_name_extractor_pairs

np.random.seed(1234)

_TEST_FILE_NAME = 'AHN3.ply'
_TEST_NEIGHBORHOODS_FILE_NAME = 'AHN3_1000_random_neighbors.json'
_TEST_DATA_SOURCE = 'testdata'

_CYLINDER = InfiniteCylinder(4)
_PC_260807 = load(os.path.join(_TEST_DATA_SOURCE, _TEST_FILE_NAME))
_PC_1000 = copy_point_cloud(_PC_260807, array_mask=(
    np.random.choice(range(len(_PC_260807[point]['x']['data'])), size=1000, replace=False)))
_PC_10 = copy_point_cloud(_PC_260807, array_mask=(
    np.random.choice(range(len(_PC_260807[point]['x']['data'])), size=10, replace=False)))
_1000_NEIGHBORHOODS_IN_260807 = list(compute_neighborhoods(_PC_260807, _PC_1000, _CYLINDER, sample_size=500))
_10_NEIGHBORHOODS_IN_260807 = list(compute_neighborhoods(_PC_260807, _PC_10, _CYLINDER, sample_size=500))
_260807_NEIGHBORHOODS_IN_10 = list(compute_neighborhoods(_PC_10, _PC_260807, _CYLINDER, sample_size=500))

features_by_name = create_default_feature_map()
feature_names = [name for name in features_by_name]


def test_no_duplicate_feature_registrations():
    pairs = _create_name_extractor_pairs()
    for name, _ in pairs:
        matches = [extractor for extractor_name, extractor in pairs if extractor_name is name]
        np.testing.assert_equal(len(matches), 1,
                                'Duplicate registrations for key "{}" by extractors: {}'.format(name, matches))


@pytest.mark.parametrize("feature", feature_names)
def test_completeTile_consistentOutput(feature):
    target_point_cloud = copy_point_cloud(_PC_1000)
    compute_features(copy_point_cloud(_PC_260807), _1000_NEIGHBORHOODS_IN_260807, target_point_cloud, [feature],
                     volume=_CYLINDER)
    _assert_consistent_attribute_length(target_point_cloud)


@pytest.mark.parametrize("feature", feature_names)
def test_manyTargets_consistentOutput(feature):
    target_point_cloud = copy_point_cloud(_PC_260807)
    compute_features(copy_point_cloud(_PC_10), _260807_NEIGHBORHOODS_IN_10, target_point_cloud, [feature],
                     volume=_CYLINDER)
    _assert_consistent_attribute_length(target_point_cloud)


@pytest.mark.parametrize("feature", feature_names)
def test_xAllZeros_consistentOutput(feature):
    n = 10
    pc = _create_point_cloud(x=0, n=n)
    compute_features(pc, [[] for _ in range(n)], pc, [feature], volume=_CYLINDER)
    _assert_consistent_attribute_length(pc)


@pytest.mark.parametrize("feature", feature_names)
def test_yAllZeros_consistentOutput(feature):
    n = 10
    pc = _create_point_cloud(y=0, n=n)
    compute_features(pc, [[] for _ in range(n)], pc, [feature], volume=_CYLINDER)
    _assert_consistent_attribute_length(pc)


@pytest.mark.parametrize("feature", feature_names)
def test_zAllZeros_consistentOutput(feature):
    n = 10
    pc = _create_point_cloud(z=0, n=n)
    compute_features(pc, [[] for _ in range(n)], pc, [feature], volume=_CYLINDER)
    _assert_consistent_attribute_length(pc)


@pytest.mark.parametrize("feature", feature_names)
def test_zeroPoints_consistentOutput(feature):
    n = 0
    pc = _create_point_cloud(n=n)
    compute_features(pc, [[] for _ in range(n)], pc, [feature], volume=_CYLINDER)
    _assert_consistent_attribute_length(pc)


@pytest.mark.parametrize("feature", feature_names)
def test_zeroNeighbors_consistentOutput(feature):
    _assert_consistent_output_with_n_neighbors(feature, 0)


@pytest.mark.parametrize("feature", feature_names)
def test_oneNeighbor_consistentOutput(feature):
    _assert_consistent_output_with_n_neighbors(feature, 1)


@pytest.mark.parametrize("feature", feature_names)
def test_twoNeighbors_consistentOutput(feature):
    _assert_consistent_output_with_n_neighbors(feature, 2)


def _assert_consistent_output_with_n_neighbors(feature, n_neighbors):
    n_points = 10
    pc = _create_point_cloud(n=n_points)
    compute_features(pc, [range(n_neighbors) for _ in range(n_points)], pc, [feature], volume=_CYLINDER)
    _assert_consistent_attribute_length(pc)


@pytest.mark.parametrize("feature", feature_names)
def test_inputNotChanged(feature):
    original_environment = _PC_260807
    environment = copy_point_cloud(original_environment)
    original_targets = _PC_10
    targets = copy_point_cloud(original_targets)
    original_neighborhoods = _10_NEIGHBORHOODS_IN_260807
    neighborhoods = [[e for e in l] for l in original_neighborhoods]

    compute_features(environment, neighborhoods, targets, [feature], volume=_CYLINDER)

    _assert_attributes_not_changed(original_environment, environment)
    _assert_attributes_not_changed(original_targets, targets)
    assert json.dumps(original_neighborhoods) == json.dumps(neighborhoods)


def _create_point_cloud(x=None, y=None, z=None, norm_z=None, intensity=None, n=10):
    tag = GROUND_TAGS[0]
    pc = {point: {'x': _create_attribute(n, fill_value=x),
                  'y': _create_attribute(n,fill_value=y),
                  'z': _create_attribute(n,fill_value=z),
                  keys.normalized_height: _create_attribute(n,fill_value=norm_z),
                  keys.intensity: _create_attribute(n,fill_value=intensity),
                  'raw_classification': {'data': np.array([i if i % 2 == 0 else tag for i in range(n)]),
                                         'type': 'float'}}}
    return pc


def _create_attribute(n_points, fill_value=None):
    attribute_data = np.array([fill_value if fill_value is not None else i for i in range(n_points)])
    return {'data': attribute_data, 'type': 'float'}


def _assert_attributes_not_changed(original_point_cloud, new_point_cloud):
    for attribute in original_point_cloud[point]:
        np.testing.assert_array_almost_equal(new_point_cloud[point][attribute]['data'],
                                             original_point_cloud[point][attribute]['data'])


def _assert_consistent_attribute_length(target_point_cloud):
    n_elements = len(target_point_cloud[point]['x']['data'])
    for key in target_point_cloud[point]:
        assert n_elements == len(target_point_cloud[point][key]['data'])



"""Test1 feature extractor."""
import numpy as np

from laserchicken import utils
from laserchicken.feature_extractor.base_feature_extractor import FeatureExtractor


class Test1FeatureExtractor(FeatureExtractor):
    @classmethod
    def requires(cls):
        return []

    @classmethod
    def provides(cls):
        return ['test1_a', 'test1_b']

    def extract(self, point_cloud, neighborhoods, target_point_cloud, target_indices, volume_description):
        return np.array([self._extract_one(target_point_cloud, target_index) for target_index in target_indices]).T

    def _extract_one(self, target_point_cloud, target_index):
        x, y, z = utils.get_point(target_point_cloud, target_index)
        return [0.5 * z, 1.5 * z]



"""Functions for loading and saving a pointcloud from/to file."""
from __future__ import print_function

from colorama import Fore

from . import ToolException
from laserchicken import load, export


def _load(filename):
    """Load point cloud from filename using reader."""
    print("Reading the input file", end='')
    point_cloud = load(filename)
    print(Fore.GREEN + "  [DONE]")
    return point_cloud


def _save(point_cloud, filename, writer=None):
    """Save point cloud to filename using writer."""
    print("File will be saved as {}".format(filename))
    print("Saving, please wait...", end='')
    try:
        export(point_cloud, filename)
    except Exception as exc:
        print(Fore.RED + "  [ERROR]")
        print("An exception of type {} occurred. Arguments:\n{!r}".format(type(exc).__name__, exc.args))
        raise ToolException("Conversion has failed! \nCheck '{}' in laserchicken module.".format(writer))

    print(Fore.GREEN + "  [DONE]")



import ast
import json
import numpy as np

from json.decoder import JSONDecodeError
from struct import unpack, calcsize

from laserchicken.io.utils import convert_to_short_type, convert_to_single_character_type


def read(path):
    """
    Read point cloud data from a ply file.
    :param path: path to the ply file
    :return: dictionary containing the point cloud data
    """
    # check whether file is in ascii/binary format
    is_binary = _is_ply_binary(path)

    # read file content
    with open(path, ''.join(['r', 'b' if is_binary else ''])) as ply:
        first_line = _read_header_line(ply, is_binary)

        if 'ply' not in first_line:
            raise ValueError('Not a valid ply file: {}'.format(path))

        index, format = _read_header(ply, is_binary)
        return {block['type']: _read_block(block, ply, format) for block in index}


def _is_ply_binary(path, is_binary=False):
    is_valid = True
    with open(path, ''.join(['r', 'b' if is_binary else ''])) as ply:
        try:
            _ = _read_header_line(ply, is_binary)
        except UnicodeDecodeError:
            if is_binary:
                is_valid = False
            else:
                return _is_ply_binary(path, is_binary=True)
    if not is_valid:
        raise ValueError('Not a valid ply file: {}'.format(path))
    return is_binary


def _read_header(ply, is_binary=False):
    index = []
    comments = []
    line = _read_header_line(ply, is_binary)
    format = line.split()[1]
    while line.strip() != 'end_header':
        if line.startswith('element'):
            element_type = line.split()[1]
            number_of_elements = int(line.split()[2])
            current_properties = []
            index.append(
                {'type': element_type, 'number_of_elements': number_of_elements, 'properties': current_properties})

        if line.startswith('property'):
            property_type, property_name = line[9:].rstrip().split(' ')
            current_properties.append({'type': property_type, 'name': property_name})

        if line.startswith('comment'):
            comment_line = line.rstrip().split(' ', 1)[1]
            comments.append(comment_line)

        line = _read_header_line(ply, is_binary)

    index.append({'type': 'log', 'log': (_read_log(comments))})
    return index, format


def _read_header_line(ply, is_binary=False):
    line = ply.readline()
    if not is_binary:
        return line
    else:
        return line.decode('utf-8')


def _read_log(comments):
    try:
        log = json.loads(' '.join(comments)) if comments else []
    except JSONDecodeError:
        try:
            # legacy: comments for laserchicken < 0.4.0
            log = ast.literal_eval(' '.join(comments)) if comments else []
        except SyntaxError:  # Log can't be read. Maybe a ply file with 'regular' comments and no log.
            log = []
    return log


def _read_block(block, ply_body, format='ascii'):
    if block['type'] == 'log':
        return block['log']
    else:
        properties, property_names = _get_properties(block)
        block_type = block['type']
        number_of_elements = block['number_of_elements']

        if format == 'ascii':
            _read_elements_ascii(ply_body, properties, property_names, block_type, number_of_elements)
        else:
            _read_elements_binary(ply_body, properties, property_names, block_type, number_of_elements, format)

        return properties


def _cast(value, value_type):
    dtype = np.dtype(value_type)
    return dtype.type(value)


def _read_elements_ascii(ply_body, properties, property_names, block_type, number_of_elements):
    for i in range(number_of_elements):
        line = ply_body.readline()
        values = line.split()
        if len(values) != len(property_names):
            raise ValueError('Error reading line {} of {} list.'.format(i, block_type))
        for p, property_name in enumerate(property_names):
            property_type = properties[property_name]['type']
            value = _cast(values[p], property_type)
            properties[property_name]['data'][i] = value


def _read_elements_binary(ply_body, properties, property_names, block_type, number_of_elements, bin_format):
    if bin_format == 'binary_little_endian':
        byte_ordering = '<'
    elif bin_format == 'binary_big_endian':
        byte_ordering = '>'
    else:
        raise ValueError('Unable to read: unknown binary format {}.'.format(bin_format))

    format_list = [convert_to_single_character_type(properties[name]['type'])
                   for name in property_names]
    format = ''.join([byte_ordering] + format_list)
    buffer_size = calcsize(format)
    for i in range(number_of_elements):
        buffer = ply_body.read(buffer_size)
        try:
            values = unpack(format, buffer)
        except Exception as err:
            raise ValueError('{}; error at line {} of {} list.'.format(str(err), i, block_type))
        for p, property_name in enumerate(property_names):
            properties[property_name]['data'][i] = values[p]


def _get_properties(block):
    properties = {}
    property_names = []
    for prop in block['properties']:
        dtype = np.dtype(convert_to_short_type(prop['type'], use_ply_implicit=True))
        property_name = prop['name']
        properties[property_name] = {'type': dtype.name, 'data': np.zeros(block['number_of_elements'], dtype)}
        property_names.append(property_name)
    return properties, property_names



"""Test2 and Test3 feature extractors."""
import numpy as np

from laserchicken import utils
from laserchicken.feature_extractor.base_feature_extractor import FeatureExtractor


class Test2FeatureExtractor(FeatureExtractor):
    @classmethod
    def requires(cls):
        return ['test1_b']

    @classmethod
    def provides(cls):
        return ['test2_a', 'test2_b', 'test2_c']

    def extract(self, point_cloud, neighborhoods, target_point_cloud, target_indices, volume_description):
        return np.array([self._extract_one(target_point_cloud, target_index) for target_index in target_indices]).T

    def _extract_one(self, target_point_cloud, target_index):
        t1b = utils.get_attribute_value(target_point_cloud, target_index, self.requires()[0])
        x, y, z = utils.get_point(target_point_cloud, target_index)
        return [x + t1b, y + t1b, z + t1b]  # x + 3z/2, y + 3z/2, 5z/2


class Test3FeatureExtractor(FeatureExtractor):
    @classmethod
    def requires(cls):
        return ['test1_a', 'test2_c']

    @classmethod
    def provides(cls):
        return ['test3_a']

    def extract(self, point_cloud, neighborhoods, target_point_cloud, target_indices, volume_description):
        return np.array([self._extract_one(target_point_cloud, target_index) for target_index in target_indices]).T

    def _extract_one(self, target_point_cloud, target_index):
        t1a, t2c = utils.get_features(target_point_cloud, self.requires(), target_index)
        x, y, z = utils.get_point(target_point_cloud, target_index)
        return t2c - t1a - z  # this should be: 2 z


class TestVectorizedFeatureExtractor(FeatureExtractor):
    @classmethod
    def requires(cls):
        return []

    @classmethod
    def provides(cls):
        return ['vectorized1', 'vectorized2']

    def extract(self, sourcepc, neighborhood, targetpc, targetindex, volume):
        x = utils.get_features(targetpc, ['x'], targetindex)
        x1 = np.array(list(x))[0]
        return x1, x1



from unittest import TestCase

from numpy.testing import assert_equal, assert_almost_equal

from laserchicken.volume_specification import Sphere, InfiniteCylinder, Cell, Cube


class VolumeTests(TestCase):
    @staticmethod
    def test_sphere_correctType():
        assert_equal(Sphere(1).get_type(), Sphere.TYPE)

    @staticmethod
    def test_sphere_calculateVolume():
        assert_almost_equal(Sphere(2).calculate_volume(), 33.510321638)

    @staticmethod
    def test_infiniteCylinder_type():
        assert_equal(InfiniteCylinder(2).get_type(), InfiniteCylinder.TYPE)

    @staticmethod
    def test_infiniteCylinder_calculateArea():
        assert_almost_equal(InfiniteCylinder(2).calculate_base_area(), 12.56637061436)

    @staticmethod
    def test_cell_type():
        assert_equal(Cell(2).get_type(), Cell.TYPE)

    @staticmethod
    def test_cell_calculateArea():
        assert_almost_equal(Cell(2).calculate_base_area(), 4.0)

    @staticmethod
    def test_cube_correctType():
        assert_equal(Cube(1).get_type(), Cube.TYPE)

    @staticmethod
    def test_cube_calculateVolume():
        assert_almost_equal(Cube(2).calculate_volume(), 8.0)



"""Pulse penetration ratio and density absolute mean calculations.

See https://github.com/eEcoLiDAR/eEcoLiDAR/issues/23.
"""

import numpy as np

from laserchicken.feature_extractor.base_feature_extractor import FeatureExtractor
from laserchicken.keys import point, normalized_height

# classification according to
# http://www.asprs.org/wp-content/uploads/2010/12/LAS_1-4_R6.pdf
GROUND_TAGS = [2]


def _is_ground(i, point_cloud):
    return point_cloud[point]['raw_classification']["data"][i] in GROUND_TAGS


class DensityAbsoluteMeanFeatureExtractor(FeatureExtractor):
    """Feature extractor for the point density."""
    def __init__(self, data_key='z'):
        self.data_key = data_key

    @classmethod
    def requires(cls):
        return []

    def provides(self):
        return ['density_absolute_mean_' + self.data_key]

    def extract(self, point_cloud, neighborhoods, target_point_cloud, target_indices, volume_description):
        return [self._extract_one(point_cloud, neighborhood) for neighborhood in neighborhoods]

    def _extract_one(self, point_cloud, neighborhood):
        if 'raw_classification' not in point_cloud[point]:
            raise ValueError(
                'Missing raw_classification attribute which is necessary for calculating density_absolute_mean.')

        non_ground_indices = [i for i in neighborhood if not _is_ground(i, point_cloud)]
        density_absolute_mean_z = self._get_density_absolute_mean(non_ground_indices, point_cloud)

        return density_absolute_mean_z

    @staticmethod
    def _get_ground_indices(point_cloud, ground_tags):
        index_grd = []
        for ipt, c in enumerate(point_cloud):
            if c in ground_tags:
                index_grd.append(ipt)
        return index_grd

    def _get_density_absolute_mean(self, non_ground_indices, source_point_cloud):
        n_non_ground = len(non_ground_indices)
        data_non_ground = source_point_cloud[point][self.data_key]["data"][non_ground_indices]
        if n_non_ground == 0:
            density_absolute_mean = 0.
        else:
            density_absolute_mean = float(
                len(data_non_ground[data_non_ground > np.mean(data_non_ground)])) / n_non_ground * 100.
        return density_absolute_mean

    def get_params(self):
        return ()



""" IO handler for PLY file format """
from laserchicken.io.base_io_handler import IOHandler
from laserchicken.io.ply_read import read
from laserchicken.io.ply_write import write


class PLYHandler(IOHandler):
    """ Class for IO of point-cloud data in PLY file format """

    def read(self):
        return read(self.path)

    def write(self, point_cloud, *args, **kwargs):
        write(point_cloud, self.path, *args, **kwargs)



import pytest
from click.testing import CliRunner
from . import cli


@pytest.fixture
def runner():
    return CliRunner()


# def test_cli(runner):
#     result = runner.invoke(cli.main)
#     assert result.exit_code == 0
#     assert not result.exception
#     assert result.output.strip() == 'Hello.'
#
#
# def test_cli_with_option(runner):
#     result = runner.invoke(cli.main, ['--', './testdata/AHN2.las'])
#     assert not result.exception
#     assert result.exit_code == 0
#     assert result.output.strip() == ''
#
#
# def test_cli_with_arg(runner):
#     result = runner.invoke(cli.main, ['Chicken'])
#     assert result.exit_code == 0
#     assert not result.exception
#     assert result.output.strip() == ''



import os
import random
import unittest

import numpy as np

from laserchicken import keys, load, utils
from laserchicken.compute_neighbors import compute_neighborhoods
from laserchicken.feature_extractor.pulse_penetration_feature_extractor import PulsePenetrationFeatureExtractor
from laserchicken.volume_specification import InfiniteCylinder


class TestPulsePenetrationFeatureExtractorArtificialData(unittest.TestCase):
    """Test the pulse extractor on artificial data."""

    def test_pulse(self):
        """Pulse extractor on artificial data should yield expected feature values."""
        extractor = PulsePenetrationFeatureExtractor()
        pp_ratio = extractor.extract(self.point_cloud, [self.neighborhood], None, None, None)[0]
        self.assertEqual(pp_ratio, self.expected_pp_ratio)

    def _set_plane_data(self):
        """Create two planes of ground point at z = +- 0.1."""
        n_points = 10
        pos = np.linspace(-self.radius, self.radius, n_points)
        self.points_per_plane = 2 * n_points ** 2
        for z in [-0.1, 0.1]:
            for x in pos:
                for y in pos:
                    self.xyz.append([x, y, z])
                    self.pt_type.append(2)

    def _set_sphere_data(self):
        """Create a sphere of vegetation point centered at 0."""
        nteta, nphi = 11, 11
        self.npt_sphere = nteta * nphi
        teta = np.linspace(0.1, 2 * np.pi, nteta)
        phi = np.linspace(0.1, np.pi, nphi)
        r = self.radius
        for t in teta:
            for p in phi:
                x = r * np.cos(t) * np.sin(p)
                y = r * np.sin(t) * np.sin(p)
                z = r * np.cos(p)
                self.xyz.append([x, y, z])
                self.pt_type.append(np.random.randint(3, 6))

    def _get_pc(self):
        return {keys.point: {'x': {'type': 'double', 'data': self.xyz[:, 0]},
                             'y': {'type': 'double', 'data': self.xyz[:, 1]},
                             'z': {'type': 'double', 'data': self.xyz[:, 2]},
                             'raw_classification': {'type': 'int', 'data': self.pt_type}}}, len(self.xyz)

    def setUp(self):
        """Set up the data for the test."""
        # set up the point cloud
        self.radius = 0.5
        self.xyz, self.pt_type = [], []
        self._set_sphere_data()
        self._set_plane_data()
        self.xyz = np.array(self.xyz)
        self.point_cloud, n_points = self._get_pc()

        # get the entire neighborhood
        self.neighborhood = list(range(n_points))

        # theo val
        self.expected_pp_ratio = float(self.points_per_plane) / n_points


class TestPulsePenetratioFeatureExtractorRealData(unittest.TestCase):
    """Test the pulse extractor on real data and make sure it doesn't crash."""
    _test_file_name = 'AHN3.las'
    _test_data_source = 'testdata'
    point_cloud = None
    target_point_cloud = None
    cyl = None
    target_point_cloud_index = None

    def test_valid(self):
        """Compute the echo ratio for a sphere/cylinder at different target points without crashing."""
        # read the data
        self.point_cloud = load(os.path.join(self._test_data_source, self._test_file_name))

        # get the target point clouds
        random.seed(102938482634)
        self.target_point_cloud = self._get_random_targets()
        self.target_point_cloud_index = 0

        # volume descriptions
        radius = 0.5
        self.cyl = InfiniteCylinder(radius)
        neighborhoods = compute_neighborhoods(self.point_cloud, self.target_point_cloud, self.cyl)

        # extractor
        extractor = PulsePenetrationFeatureExtractor()
        extractor.extract(self.point_cloud, neighborhoods, None, None, None)

    def _get_random_targets(self):
        """Get a random target pc."""
        num_all_pc_points = len(self.point_cloud[keys.point]["x"]["data"])
        rand_indices = [random.randint(0, num_all_pc_points)
                        for _ in range(20)]
        return utils.copy_point_cloud(self.point_cloud, rand_indices)


if __name__ == '__main__':
    unittest.main()



import sys
import math
import numpy as np
from scipy.spatial import cKDTree
from psutil import virtual_memory

from laserchicken import utils, kd_tree
from laserchicken.volume_specification import Sphere, InfiniteCylinder, Cell, Cube
from laserchicken.keys import point


def frange(x_value, y_value, jump):
    while x_value < y_value:
        yield x_value
        x_value += jump


MEMORY_THRESHOLD = 0.5
POINT_CLOUD_DIST = 10


def compute_cylinder_neighborhood(environment_pc, target_pc, radius):
    """Find the indices of points within a cylindrical neighbourhood (using KD Tree) for a given point of a target
    point cloud among the points from an environment point cloud.

    :param environment_pc: environment point cloud
    :param target_pc: point cloud that contains the points at which neighborhoods are to be calculated
    :param radius: search radius for neighbors
    :return: indices of neighboring points from the environment point cloud for each target point
             the returned indices also contains the index of the target point.
    """
    avg_points_cyl = (radius * radius * math.pi) * POINT_CLOUD_DIST
    target_points = target_pc[point]['x']['data']

    if len(environment_pc[point]['x']['data']) == 0:
        for _ in target_points:
            yield []
        return

    cyl_size = avg_points_cyl * np.size(target_points) * sys.getsizeof(int)
    mem_size = virtual_memory().total

    print("Cylinder size in Bytes: %s" % cyl_size)
    print("Memory size in Bytes: %s" % mem_size)

    if cyl_size > mem_size * MEMORY_THRESHOLD:
        y = target_pc[point]['y']['data']

        num_points = int(math.floor(mem_size * MEMORY_THRESHOLD / (avg_points_cyl * sys.getsizeof(int))))
        print("Number of points: %d" % num_points)

        env_tree = kd_tree.get_kdtree_for_pc(environment_pc)

        for range_start in range(0, np.size(target_points), num_points):
            range_end = min(range_start + num_points, np.size(target_points))
            box_points = np.column_stack((target_points[range_start:range_end], y[range_start:range_end]))
            target_box_tree = cKDTree(box_points, compact_nodes=False, balanced_tree=False)
            for neighborhood in target_box_tree.query_ball_tree(env_tree, radius):
                yield neighborhood

    else:
        print("Start tree creation")
        env_tree = kd_tree.get_kdtree_for_pc(environment_pc)
        print("Done with env tree creation")
        target_tree = kd_tree.get_kdtree_for_pc(target_pc)
        print("Done with target tree creation")
        for neighborhood in target_tree.query_ball_tree(env_tree, radius):
            yield neighborhood


def compute_sphere_neighborhood(environment_pc, target_pc, radius):
    """
    Find the indices of points within a spherical neighbourhood for a given point of a target point cloud among the
    points from an environment point cloud.

    :param environment_pc: environment point cloud
    :param target_pc: point cloud that contains the points at which neighborhoods are to be calculated
    :param radius: search radius for neighbors
    :return: indices of neighboring points from the environment point cloud for each target point
    """
    neighborhoods = compute_cylinder_neighborhood(environment_pc, target_pc, radius)
    for target_i, neighborhood_indices in enumerate(neighborhoods):
        target_x, target_y, target_z = utils.get_point(target_pc, target_i)
        result_indices = []
        for neighbor_j in neighborhood_indices:
            env_x, env_y, env_z = utils.get_point(environment_pc, neighbor_j)
            if abs(target_z - env_z) > radius:
                continue
            if (env_x - target_x) ** 2 + (env_y - target_y) ** 2 + (env_z - target_z) ** 2 <= radius ** 2:
                result_indices.append(neighbor_j)
        yield result_indices


def compute_cell_neighborhood(environment_pc, target_pc, side_length):
    """
    Find the indices of points within a square neighbourhood for a given point of a target point cloud among the
    points from an environment point cloud.

    :param environment_pc: environment point cloud
    :param target_pc: point cloud that contains the points at which neighborhoods are to be calculated
    :param side_length: search radius for neighbors
    :return: indices of neighboring points from the environment point cloud for each target point
    """

    max_radius = 0.5 * math.sqrt((side_length ** 2) + (side_length ** 2))

    neighborhoods = compute_cylinder_neighborhood(
        environment_pc, target_pc, max_radius)

    for target_i, neighborhood_indices in enumerate(neighborhoods):
        target_x, target_y, _ = utils.get_point(target_pc, target_i)
        result_indices = []
        for neighbor_j in neighborhood_indices:
            env_x, env_y, _ = utils.get_point(environment_pc, neighbor_j)
            if ((abs(target_x - env_x)) > 0.5 * side_length) or ((abs(target_y - env_y)) > 0.5 * side_length):
                continue
            else:
                result_indices.append(neighbor_j)
        yield result_indices


def compute_cube_neighborhood(environment_pc, target_pc, side_length):
    """
    Find the indices of points within a square neighbourhood for a given point of a target point cloud among the
    points from an environment point cloud.

    :param environment_pc: environment point cloud
    :param target_pc: point cloud that contains the points at which neighborhoods are to be calculated
    :param side_length: search radius for neighbors
    :return: indices of neighboring points from the environment point cloud for each target point
    """

    neighborhoods = compute_cell_neighborhood(
        environment_pc, target_pc, side_length)

    for target_i, neighborhood_indices in enumerate(neighborhoods):
        _, _, target_z = utils.get_point(target_pc, target_i)
        result_indices = []
        for neighbor_j in neighborhood_indices:
            _, _, env_z = utils.get_point(environment_pc, neighbor_j)
            if abs(target_z - env_z) > side_length:
                continue
            else:
                result_indices.append(neighbor_j)
        yield result_indices



def compute_neighborhoods(env_pc, target_pc, volume_description, sample_size=None):
    """
    Find a subset of points in a neighbourhood in the environment point cloud for each point in a target point cloud.

    :param env_pc: environment point cloud
    :param target_pc: point cloud that contains the points at which neighborhoods are to be calculated
    :param volume_description: volume object that describes the shape and size of the search volume
    :param sample_size: maximum number of neighbors returned per target point; if None (default), all are returned
    :return: indices of neighboring points from the environment point cloud for each target point
    """
    volume_type = volume_description.get_type()

    if volume_type == Cell.TYPE:
        neighborhoods = compute_cell_neighborhood(env_pc, target_pc, volume_description.side_length)
    elif volume_type == Cube.TYPE:
        neighborhoods = compute_cube_neighborhood(env_pc, target_pc, volume_description.side_length)
    elif volume_type == Sphere.TYPE:
        neighborhoods = compute_sphere_neighborhood(env_pc, target_pc, volume_description.radius)
    elif volume_type == InfiniteCylinder.TYPE:
        neighborhoods = compute_cylinder_neighborhood(env_pc, target_pc, volume_description.radius)
    else:
        raise ValueError(
            'Neighborhood computation error because volume type "{}" is unknown.'.format(volume_type))

    return _subsample_if_necessary(neighborhoods, sample_size)


def _subsample_if_necessary(neighborhoods, sample_size):
    if sample_size:
        sampled = (_subsample(neighborhood, sample_size) if len(neighborhood) > sample_size else neighborhood for
                   neighborhood in neighborhoods)
        return sampled
    else:
        return neighborhoods


def _subsample(neighborhood, sample_size=None):
    return list(np.random.choice(neighborhood, sample_size, replace=False))



import scipy.stats.stats as stats

from laserchicken.feature_extractor.base_feature_extractor import FeatureExtractor
from laserchicken.keys import point


class PercentileFeatureExtractor(FeatureExtractor):
    """Height percentiles feature extractor class."""
    def __init__(self, percentile=50, data_key='z'):
        self.percentile = percentile
        self.data_key = data_key

    @classmethod
    def requires(cls):
        return []

    def provides(self):
        return [self.generate_feature_name(self.percentile)]

    def generate_feature_name(self, percentile):
        return 'perc_{}_{}'.format(percentile, self.data_key)

    def extract(self, point_cloud, neighborhoods, target_point_cloud, target_indices, volume_description):
        return [self._extract_one(point_cloud, neighborhood) for neighborhood in neighborhoods]

    def _extract_one(self, point_cloud, neighborhood):
        source_data = point_cloud[point][self.data_key]['data'][neighborhood]
        return stats.scoreatpercentile(source_data, self.percentile)

    def get_params(self):
        return ()



from ._version import __version__

from laserchicken.compute_neighbors import compute_neighborhoods
from laserchicken.feature_extractor.feature_extraction import compute_features, register_new_feature_extractor
from laserchicken.io.load import load
from laserchicken.io.export import export
from laserchicken.build_volume import build_volume



import json
import numpy as np

from struct import pack

from laserchicken import keys
from laserchicken.io.utils import convert_to_single_character_type, select_valid_attributes


def write(point_cloud, path, attributes='all', is_binary=False):
    """
    Write the point cloud to a ply-file.
    :param point_cloud:
    :param attributes: write only the specified attributes ('all' to write all of them)
    :param path:
    :param is_binary:
    :return:
    """
    if keys.point in point_cloud:
        attributes = select_valid_attributes([attr for attr in point_cloud[keys.point].keys()], attributes)
    with open(path, 'w') as ply:
        _write_header(point_cloud, attributes, ply, is_binary)
    with open(path, ''.join(['a', 'b' if is_binary else ''])) as ply:
        _write_data(point_cloud, attributes, ply, is_binary)


def _write_header(point_cloud, attributes, ply, is_binary=False):
    ply.write("ply" + '\n')
    if is_binary:
        file_type = "binary_little_endian"
    else:
        file_type = "ascii"
    ply.write("format %s 1.0\n" % file_type)
    _write_comment(point_cloud, ply)
    for elem_name in _get_ordered_elements(point_cloud.keys()):
        get_num_elements = (lambda d: len(d["x"].get("data", []))) if elem_name == keys.point else None
        _write_header_elements(point_cloud, attributes, ply, elem_name, get_num_elements=get_num_elements)
    ply.write("end_header" + '\n')


def _write_data(pc, attributes, ply, is_binary=False):
    for elem_name in _get_ordered_elements(pc.keys()):
        if elem_name == keys.point:
            props = attributes
        else:
            props = pc[elem_name].keys()
        props = _get_ordered_properties(elem_name, props)
        num_elements = len(pc[elem_name]["x"].get("data", [])) if elem_name == keys.point else 1
        for i in range(num_elements):
            line_elements = []
            line_types = []
            for prop in props:
                data_values, data_type = _get_data_and_type(pc[elem_name][prop])
                if isinstance(data_values, np.ndarray):
                    line_elements.append(_format_ply(data_values[i], is_binary))
                else:
                    if i != 0:
                        raise Exception("Scalar quantity does not have element at index %d" % i)
                    line_elements.append(_format_ply(data_values, is_binary))
                line_types.append(convert_to_single_character_type(data_type, use_ply_implicit=True))
            if is_binary:
                # add leading byte order character (< for little endian)
                format = ''.join(["<"] + line_types)
                line = pack(format, *line_elements)
            else:
                line = " ".join(line_elements) + "\n"
            ply.write(line)


def _get_data_and_type(attribute):
    return attribute["data"], attribute["type"]


def _format_ply(obj, is_binary):
    if is_binary:
        return obj
    else:
        return str(obj)


def _get_ordered_elements(elem_names):
    if keys.point in elem_names:
        return [keys.point] + sorted([e for e in elem_names if e not in [keys.point, keys.provenance]])
    else:
        return sorted([e for e in elem_names if e not in [keys.point, keys.provenance]])


def _get_ordered_properties(elem_name, prop_list):
    if elem_name == keys.point:
        return ['x', 'y', 'z'] + [k for k in sorted(prop_list) if k not in ['x', 'y', 'z']]
    else:
        return sorted(prop_list)


def _write_comment(pc, ply):
    log = pc.get(keys.provenance, [])
    if not any(log):
        return

    head = 'comment [\n'
    tail = 'comment ]\n'
    formatted_entries = ',\n'.join(['comment ' + json.dumps(entry, sort_keys=True) for entry in log]) + '\n'
    ply.write(head + formatted_entries + tail)


def _write_header_elements(pc, attributes, ply, element_name, get_num_elements=None):
    if element_name in pc:
        num_elements = get_num_elements(pc[element_name]) if get_num_elements else 1
        ply.write("element %s %d\n" % (element_name, num_elements))
        if element_name == keys.point:
            key_list = attributes
        else:
            key_list = pc[element_name].keys()
        key_list = _get_ordered_properties(element_name, key_list)
        for key in key_list:
            property_type = pc[element_name][key]["type"]
            property_tuple = ("property", property_type, key)
            ply.write(" ".join(property_tuple) + '\n')



import numpy as np
import scipy.stats.stats as stat

from laserchicken.feature_extractor.base_feature_extractor import FeatureExtractor
from laserchicken.keys import point


class SkewFeatureExtractor(FeatureExtractor):
    """Calculates the skew on the z axis."""
    def __init__(self, data_key='z'):
        self.data_key = data_key

    @classmethod
    def requires(cls):
        return []

    def provides(self):
        return ['skew_' + self.data_key]

    def extract(self, point_cloud, neighborhoods, target_point_cloud, target_indices, volume_description):
        return [self._extract_one(point_cloud, neighborhood) for neighborhood in neighborhoods]

    def _extract_one(self, point_cloud, neighborhood):
        if neighborhood:
            source_data = point_cloud[point][self.data_key]['data'][neighborhood]
            skew = stat.skew(source_data)
        else:
            skew = np.NaN
        return skew



import unittest

import numpy as np

from laserchicken import load, keys
from laserchicken.feature_extractor.var_feature_extractor import VarianceFeatureExtractor
from laserchicken.test_tools import create_point_cloud


class TestVariationZFeatureExtractor(unittest.TestCase):
    def test_height_stats(self):
        pc_in = load("testdata/AHN2.las")
        neighborhood = [89664, 23893, 30638, 128795, 62052, 174453, 29129, 17127, 128215, 29667, 116156, 119157, 98591,
                        7018,
                        61494, 65194, 117931, 62971, 10474, 90322]
        var_z = self.extractor.extract(pc_in, [neighborhood], None, None, None)[0]
        np.testing.assert_allclose(var_z, 1.8408359999999995)

    def test_height_stats_without_neighbors(self):
        pc_in = load("testdata/AHN2.las")
        neighborhood = []
        var_z = self.extractor.extract(pc_in, [neighborhood], pc_in, None, None)[0]
        assert np.isnan(var_z)

    def test_default_provides_correct(self):
        feature_names = self.extractor.provides()
        self.assertIn('var_z', feature_names)

    def setUp(self):
        self.extractor = VarianceFeatureExtractor()


class TestVarianceNormZFeatureExtractor(unittest.TestCase):
    def test_height_stats(self):
        x = y = np.array([0, 0, 0])
        z = np.array([2, 2, 2])
        normalized_z = np.array([3, 4, 5])
        point_cloud = create_point_cloud(x, y, z, normalized_z=normalized_z)
        neighborhood = [[0, 1, 2]]

        variance = self.extractor.extract(point_cloud, [neighborhood], None, None, None)[0]

        self.assertAlmostEqual(variance, 2 / 3)

    def test_default_provides_correct(self):
        feature_names = self.extractor.provides()
        self.assertIn('var_normalized_height', feature_names)

    def setUp(self):
        self.extractor = VarianceFeatureExtractor(data_key=keys.normalized_height)



"""Abstract feature extractor."""


class FeatureExtractor(object):
    """Abstract feature extractor class."""

    @classmethod
    def requires(cls):
        """
        Get a list of names of the point attributes that are needed for this feature extraction.

        For simple features, this could be just x, y, and z. Other features can build on again
        other features to have been computed first.

        :return: List of feature names
        """
        raise NotImplementedError(
            "Class %s doesn't implement get_requirements()" % (cls.__name__))

    @classmethod
    def provides(cls):
        """
        Get a list of names of the feature values.

        This will return as many names as the number feature values that will be returned.
        For instance, if a feature extractor returns the first 3 Eigen values, this method
        should return 3 names, for instance 'eigen_value_1', 'eigen_value_2' and 'eigen_value_3'.

        :return: List of feature names
        """
        raise NotImplementedError(
            "Class %s doesn't implement get_names()" % (cls.__name__))

    def extract(self, point_cloud, neighborhoods, target_point_cloud, target_index, volume_description):
        """
        Extract the feature value(s) of the point cloud at location of the target.

        :param point_cloud: environment (search space) point cloud
        :param neighborhoods: list of arrays of indices of points within the point_cloud argument
        :param target_point_cloud: point cloud that contains target point
        :param target_indices: list of indices of the target point in the target point cloud
        :param volume_description: volume object that describes the shape and size of the search volume
        :return: feature values
        """
        raise NotImplementedError(
            "Class %s doesn't implement extract_features()" % (self.__class__.__name__))

    def get_params(self):
        """
        Returns a tuple of parameters involved in the current feature extractor
        object. Needed for provenance.
        """
        return ()





import os
import unittest


class FromTutorial(unittest.TestCase):
    def test_tutorial_once(self):
        """This test should be identical to running all cells in the tutorial notebook once, in order."""
        from laserchicken import load
        point_cloud = load('testdata/AHN3.las')

        point_cloud

        from laserchicken.normalize import normalize
        normalize(point_cloud)

        point_cloud

        from laserchicken.filter import select_polygon
        polygon = "POLYGON(( 131963.984125 549718.375000," + \
                  " 132000.000125 549718.375000," + \
                  " 132000.000125 549797.063000," + \
                  " 131963.984125 549797.063000," + \
                  " 131963.984125 549718.375000))"
        point_cloud = select_polygon(point_cloud, polygon)

        from laserchicken.filter import select_above, select_below
        points_below_1_meter = select_below(point_cloud, 'normalized_height', 1)
        points_above_1_meter = select_above(point_cloud, 'normalized_height', 1)

        from laserchicken import compute_neighborhoods
        from laserchicken import build_volume
        targets = point_cloud
        volume = build_volume("sphere", radius=5)
        neighborhoods = compute_neighborhoods(point_cloud, targets, volume)

        from laserchicken import compute_features
        compute_features(point_cloud, neighborhoods, targets, ['std_z', 'mean_z', 'slope'], volume)

        from laserchicken import register_new_feature_extractor
        from laserchicken.feature_extractor.band_ratio_feature_extractor import BandRatioFeatureExtractor
        register_new_feature_extractor(BandRatioFeatureExtractor(None, 1, data_key='normalized_height'))
        register_new_feature_extractor(BandRatioFeatureExtractor(1, 2, data_key='normalized_height'))
        register_new_feature_extractor(BandRatioFeatureExtractor(2, None, data_key='normalized_height'))
        register_new_feature_extractor(BandRatioFeatureExtractor(None, 0, data_key='z'))

        from laserchicken.feature_extractor.feature_extraction import list_feature_names
        sorted(list_feature_names())

        cylinder = build_volume("infinite cylinder", radius=5)
        neighborhoods = compute_neighborhoods(point_cloud, targets, cylinder)
        compute_features(point_cloud, neighborhoods, targets, ['band_ratio_1<normalized_height<2'], cylinder)

        from laserchicken import export
        export(point_cloud, 'my_output.ply')

    def tearDown(self):
        os.remove('my_output.ply')


import os

from .las_handler import LASHandler as las
from .ply_handler import PLYHandler as ply

io_handlers = {
    '.ply': ply,
    '.las': las,
    '.laz': las,
}


def get_io_handler(path, mode, format=None, overwrite=False):
    """
    Return instance of format-specific IOHandler, already initialized to read or write mode.

    :param path: path where the IO operation needs to take place
    :param mode: 'r' for reading, 'w' for writing
    :param format: point-cloud file format, try to guess it from extension if not specified
    :param overwrite: if working in write mode, allow to overwrite if file exists.
    :return: instance of the IOHandler
    """
    if format is None:
        _root, format = os.path.splitext(path)
    format = format.lower()
    _check_format(format)
    io_handler = io_handlers[format]
    return io_handler(path, mode, overwrite=overwrite)


def _check_format(format):
    if format not in io_handlers:
        raise NotImplementedError(
            "File format %s unknown. Implemented formats are: %s" % (format, ', '.join(io_handlers.keys())))



"""Classes that describe volumes."""
import numpy as np

SPHERE_VOLUME_FACTOR = 4 * np.pi / 3


class Volume(object):
    """Abstract class for volume descriptions."""

    def get_type(self):
        """
        Give the type as a string like 'sphere' or 'cylinder'.

        :return: type description
        """
        raise NotImplementedError("Class {} doesn't implement get_type()".format(type(self).__name__))

    def calculate_area_or_volume(self):
        """
        Calculate the area (infinite cylinder. cell) or volume (sphere, cube) depending on the volume type.

        :return: area or volume
        """
        raise NotImplementedError("Class {} doesn't implement calculate_area_or_volume()".format(type(self).__name__))


class Sphere(Volume):
    """Mathematical sphere."""

    TYPE = 'sphere'

    def __init__(self, radius):
        self.radius = radius

    def get_type(self):
        return self.TYPE

    def calculate_volume(self):
        """
        Calculate the volume of this sphere based on its radius.

        :return: volume
        """
        return np.power(self.radius, 3) * SPHERE_VOLUME_FACTOR

    def calculate_area_or_volume(self):
        return self.calculate_volume()

class InfiniteCylinder(Volume):
    """Cylinder with base in the xy plane and height that extends infinitely to both z and -z direction."""

    TYPE = 'infinite cylinder'

    def __init__(self, radius):
        self.radius = radius

    def get_type(self):
        return self.TYPE

    def calculate_base_area(self):
        """
        Calculate the area of the base of the infinite cylinder.

        :return: area of the base
        """
        return np.power(self.radius, 2) * np.pi

    def calculate_area_or_volume(self):
        return self.calculate_base_area()


class Cell(Volume):
    """Cell that is square in xy plane and extends infinitely to both z and -z direction."""

    TYPE = 'cell'

    def __init__(self, side_length):
        self.side_length = side_length

    def get_type(self):
        return self.TYPE

    def calculate_base_area(self):
        """
        Calculate the area of the base of the cell..

        :return: area of the base
        """
        return np.power(self.side_length, 2)

    def calculate_area_or_volume(self):
        return self.calculate_base_area()


class Cube(Volume):
    """Mathematical cube."""

    TYPE = 'cube'

    def __init__(self, side_length):
        self.side_length = side_length

    def get_type(self):
        return self.TYPE

    def calculate_volume(self):
        """
        Calculate the volume of this sphere based on its radius.

        :return: volume
        """
        return np.power(self.side_length, 3)

    def calculate_area_or_volume(self):
        return self.calculate_volume()



import inspect

import laserchicken
from laserchicken.volume_specification import Volume


def create_volume_map():
    """ Generate map of volume types that are available """
    volume_map = {}
    for name, obj in inspect.getmembers(laserchicken.volume_specification):
        if inspect.isclass(obj) and issubclass(obj, Volume) and obj is not Volume:
            volume_map[obj.TYPE] = obj
    return volume_map


VOLUMES = create_volume_map()


def build_volume(volume_type, *args, **kwargs):
    """
    Return volume object from the volume name and the corresponding parameters

    Example:
    >>>> vol = build_volume('sphere', radius=5)

    :param volume_type: name corresponding to the
    :param args: optional non-keyword args to build the volume
    :param kwargs: optional keyword args to build the volume
    :return:
    """
    _volume_type = volume_type.lower()
    _verify_volume_type(_volume_type)
    volume_builder = VOLUMES[_volume_type]
    return volume_builder(*args, **kwargs)


def _verify_volume_type(volume_type):
    if volume_type not in VOLUMES.keys():
        raise ValueError('Unknown volume specified: {}. Available volumes are: {}'
                         .format(volume_type, ', '.join(VOLUMES.keys())))



"""Command line tool definitions."""
from __future__ import print_function

import os

import click
from colorama import Back, init

from . import ToolException
from .._version import __version__
from ..filter import select_above, select_below, select_polygon
from .io import _load, _save


@click.group(chain=True, invoke_without_command=True)
@click.version_option(version=__version__)
@click.argument('input_file', nargs=1, type=click.Path(exists=True, readable=True))
@click.argument('output_file', nargs=1, type=click.Path(writable=True))
def main(input_file, output_file):
    """Various commmands for selecting points from a point cloud.

    Note that commands can be chained.

    Example:

    laserchicken testdata/AHN3.las test.ply filter_below intensity 100 filter_above intensity 60
    """


@main.resultcallback()
def process_pipeline(processors, input_file, output_file):
    init(autoreset=True)
    try:
        point_cloud = _load(input_file)
        for processor in processors:
            point_cloud = processor(point_cloud)
        _save(point_cloud, output_file)
    except ToolException as exc:
        print(Back.RED + "Error: " + str(exc))


@main.command('import')
def _import():
    """Read data from input file and save to output file.

    Actually the word import can be omitted, as this is the default operation.

    Examples:

    laserchicken testdata/AHN2.las test.ply

    laserchicken testdata/AHN2.las test.ply import
    """
    return lambda point_cloud: point_cloud


@main.command('filter_below')
@click.argument('attribute')
@click.argument('threshold', type=click.FLOAT)
def _filter_below(attribute, threshold):
    """Select those points where the value of attribute is below threshold.

    Example:

    laserchicken testdata/AHN3.las test.ply filter_below intensity 100
    """
    return lambda point_cloud: select_below(point_cloud, attribute, threshold)


@main.command('filter_above')
@click.argument('attribute')
@click.argument('threshold', type=click.FLOAT)
def _filter_above(attribute, threshold):
    """Select those points where the value of attribute is above threshold.

    Example:

    laserchicken testdata/AHN3.las test.ply filter_above intensity 100
    """
    return lambda point_cloud: select_above(point_cloud, attribute, threshold)


@main.command('filter_in_polygon')
@click.argument('polygon')
def _filter_in_polygon(polygon):
    """Select those points that are inside polygon.

    Polygon can be a shape file (with extension .shp or .wkt) or a WKT string.

    Examples:

    laserchicken testdata/AHN2.las test.ply filter_in_polygon testdata/ahn2_geometries_shp/ahn2_polygon.shp

    laserchicken testdata/AHN2.las test.ply filter_in_polygon testdata/ahn2_geometries_wkt/ahn2_polygon.wkt

    laserchicken testdata/AHN2.las test.ply filter_in_polygon 'POLYGON(( 243590.0 572110.0, 243640.0 572160.0, 243700.0 572110.0, 243640.0 572060.0, 243590.0 572110.0 ))'
    """
    if os.path.isfile(polygon):
        return lambda point_cloud: select_polygon(point_cloud, polygon, read_from_file=True)
    else:
        print("polygon is not a file, assuming it is a WKT string")
        return lambda point_cloud: select_polygon(point_cloud, polygon, read_from_file=False)



import numpy as np

from laserchicken.feature_extractor.base_feature_extractor import FeatureExtractor
from laserchicken.keys import point


class MeanStdCoeffFeatureExtractor(FeatureExtractor):
    """Calculates mean, standard deviation and the ratio between the two."""
    def __init__(self, data_key='z'):
        self.data_key = data_key

    @classmethod
    def requires(cls):
        return []

    def provides(self):
        base_names = ['mean_', 'std_', 'coeff_var_']
        return [base + str(self.data_key) for base in base_names]

    def extract(self, point_cloud, neighborhoods, target_point_cloud, target_indices, volume_description):
        return np.array([self._extract_one(point_cloud, neighborhood) for neighborhood in neighborhoods]).T

    def _extract_one(self, sourcepc, neighborhood):
        if neighborhood:
            z = sourcepc[point][self.data_key]['data'][neighborhood]
            mean_z = np.mean(z)
            std_z = np.std(z)
            coeff_var_z = std_z / mean_z
        else:
            mean_z = std_z = coeff_var_z = np.NaN
        return mean_z, std_z, coeff_var_z



import unittest

import numpy as np

from laserchicken import keys
from laserchicken.feature_extractor.feature_extraction import compute_features
from laserchicken.test_tools import create_point_cloud, create_points_in_xy_grid
from laserchicken.volume_specification import InfiniteCylinder


class TestExtractSigmaZ(unittest.TestCase):
    def test_constantValues_result0(self):
        def z_constant(x, y):
            return 2

        assert_std_for_z_function_in_xy_grid(z_constant, 0)

    def test_checkered1sAnd0s_resultHalf(self):
        """Standard deviation of checker pattern of 0s and 1s should be 0.5"""

        def z_checkered(x, y):
            return ((x + y) % 2) + 2

        assert_std_for_z_function_in_xy_grid(z_checkered, 0.5)

    def test_checkered1sAnd0sPlusSkewed_resultHalf(self):
        """Standard deviation of checker pattern of 0s and 1s should be 0.5, adding a plane should not change that"""

        def z_checkered(x, y):
            return ((x + y) % 2) + 2 + x + y

        assert_std_for_z_function_in_xy_grid(z_checkered, 0.5)


def assert_std_for_z_function_in_xy_grid(z_checkered, expected):
    """Assert that the standard deviation of z values in a grid of unit x and y"""
    n_points, points = create_points_in_xy_grid(z_checkered)
    point_cloud = create_point_cloud(points[:, 0], points[:, 1], points[:, 2])
    targets = create_point_cloud([0], [0], [0])
    compute_features(point_cloud, [range(n_points)], targets, [
        'sigma_z'], InfiniteCylinder(10))
    np.testing.assert_almost_equal(
        targets[keys.point]['sigma_z']['data'][0], expected)






""" IO Handler for LAS (and compressed LAZ) file format """
import pylas

from laserchicken import keys
from laserchicken.io.base_io_handler import IOHandler
from laserchicken.io.utils import convert_to_short_type, select_valid_attributes


DEFAULT_LAS_ATTRIBUTES = {
    'x',
    'y',
    'z',
    'intensity',
    'gps_time',
    'raw_classification',
}


class LASHandler(IOHandler):
    """ Class for IO of point-cloud data in LAS file format """

    def read(self, attributes=DEFAULT_LAS_ATTRIBUTES):
        """
        Load the points from a LAS(LAZ) file into memory.

        :param attributes: list of attributes to read ('all' for all attributes in file)
        :return: point cloud data structure
        """
        file = pylas.read(self.path)
        attributes_available = [el if el not in ['X', 'Y', 'Z'] else el.lower()
                                for el in file.points.dtype.names]

        attributes = select_valid_attributes(attributes_available, attributes)

        points = {}
        for name in attributes:
            if hasattr(file, name):
                data = getattr(file, name)
                points[name] = _get_attribute(data, data.dtype.name)

        return {keys.point: points}

    def write(self, point_cloud, attributes='all', file_version='1.2', point_format_id=3):
        """
        Write point cloud to a LAS(LAZ) file.

        :param point_cloud:
        :param attributes: list of attributes to write ('all' for all attributes in point_cloud)
        :param file_version:
        :param point_format_id:
        :return:
        """
        file = pylas.create(point_format_id=point_format_id,
                            file_version=file_version)

        points = point_cloud[keys.point]
        attributes = select_valid_attributes([attr for attr in points.keys()], attributes)

        # NOTE: adding extra dims and assignment should be done in two steps,
        # some fields (e.g. raw_classification) are otherwise overwritten
        for attribute in attributes:
            data, type = _get_data_and_type(points[attribute])
            type_short = convert_to_short_type(type)
            if attribute not in 'xyz':
                # x,y,z are not there but file methods can be used to convert coords to int4
                if attribute not in file.points.dtype.names:
                    file.add_extra_dim(name=attribute, type=type_short)
            file_type_short = convert_to_short_type(getattr(file, attribute).dtype.name)
            if not file_type_short == type_short:
                raise TypeError('Data type in file does not match the one in point cloud: '
                                'for {}, {} vs {}'.format(attribute, file_type_short, type_short))

        for attribute in attributes:
            data, _ = _get_data_and_type(points[attribute])
            if data.size == 0:
                raise ValueError('Cannot write empty point-cloud!')
            else:
                setattr(file, attribute, data)

        try:
            file.write(self.path)
        except ValueError as err:
            raise ValueError('Error in writing LAS file (file_version {}, point_format_id {}). '
                             'pylas error below:\n{}'.format(file_version, point_format_id, err))


def _get_attribute(data, data_type):
    return {'type': data_type, 'data': data}


def _get_data_and_type(attribute):
    return attribute['data'], attribute['type']



import dateutil
import os
import shutil
import unittest

import numpy as np
from pytest import raises

from laserchicken import keys
from laserchicken.io.load import load


class TestReadPly(unittest.TestCase):
    _test_dir = 'TestLoad_dir'
    _test_file_name = 'example.ply'
    _test_file_without_comments_name = 'example_without_comments.ply'
    _test_file_with_invalid_comments_name = 'example_with_invalid_comments.ply'
    _las_file_name = '5points.las'
    _test_data_source = 'testdata'
    las_file_path = os.path.join(_test_dir, _las_file_name)
    test_file_path = os.path.join(_test_dir, _test_file_name)
    test_file_without_comments_path = os.path.join(_test_dir, _test_file_without_comments_name)
    test_file_with_invalid_comments_path = os.path.join(_test_dir, _test_file_with_invalid_comments_name)

    def test_nonexistentFile_error(self):
        # Catch most specific subclass of FileNotFoundException (3.6) and IOError (2.7).
        with raises(Exception):
            load('nonexistentfile.ply')

    def test_wrongFormat_error(self):
        with raises(ValueError):
            load(self.las_file_path, format='.PLY')

    def test_existentPly_noError(self):
        load(self.test_file_path)

    def test_containsPointsElement(self):
        data = load(self.test_file_path)
        self.assertIn(keys.point, data)

    def test_containsXElement(self):
        data = load(self.test_file_path)
        self.assertIn('x', data[keys.point])

    def test_rightNumberOfPoints(self):
        data = load(self.test_file_path)
        self.assertEqual(len(data[keys.point]['x']['data']), 3)

    def test_correctPoints(self):
        data = load(self.test_file_path)
        points = data[keys.point]
        point = np.array(
            [points['x']['data'][0], points['y']['data'][0], points['z']['data'][0], points['return']['data'][0]])
        np.testing.assert_allclose(point, np.array([0.11, 0.12, 0.13, 1]))

    def test_correctPointCloud(self):
        data = load(self.test_file_path)
        point_cloud = data['pointcloud']
        offset = point_cloud['offset']['data'][0]
        np.testing.assert_allclose(offset, 12.1)

    def test_correctPointCloudWithoutComments(self):
        """Missing comment section should not cause error (regression test)."""
        data = load(self.test_file_without_comments_path)
        point_cloud = data['pointcloud']
        offset = point_cloud['offset']['data'][0]
        np.testing.assert_allclose(offset, 12.1)

    def test_correctPointCloudWithInvalidComments(self):
        """Invalid comments should not cause error."""
        data = load(self.test_file_with_invalid_comments_path)
        point_cloud = data['pointcloud']
        offset = point_cloud['offset']['data'][0]
        np.testing.assert_allclose(offset, 12.1)

    def test_allLogEntriesContainAllColumns(self):
        log = load(self.test_file_path)['log']

        for entry in log:
            for key in ['time', 'module', 'parameters', 'version']:
                self.assertIn(key, entry)

    def test_correctModulesLogged(self):
        log = load(self.test_file_path)['log']

        modules = [entry['module'] for entry in log]
        # an additional 'load' is added in the log when reading
        self.assertListEqual(['load', 'filter', 'laserchicken.io.load'], modules)

    def test_correctTimesLogged(self):
        log = load(self.test_file_path)['log']

        self.assertListEqual([2018, 1, 18, 16, 1, 0, 3, 18, -1],
                             list(dateutil.parser.parse(log[0]['time']).timetuple()))
        self.assertListEqual([2018, 1, 18, 16, 3, 0, 3, 18, -1],
                             list(dateutil.parser.parse(log[1]['time']).timetuple()))

    def setUp(self):
        os.mkdir(self._test_dir)
        shutil.copyfile(os.path.join(self._test_data_source, self._test_file_name), self.test_file_path)
        shutil.copyfile(os.path.join(self._test_data_source, self._test_file_without_comments_name),
                        self.test_file_without_comments_path)
        shutil.copyfile(os.path.join(self._test_data_source, self._test_file_with_invalid_comments_name),
                        self.test_file_with_invalid_comments_path)
        shutil.copyfile(os.path.join(self._test_data_source, self._las_file_name), self.las_file_path)

    def tearDown(self):
        shutil.rmtree(self._test_dir)


class TestReadPlyBinary(TestReadPly):
    _test_file_name = 'example_little_endian.ply'
    _test_file_without_comments_name = 'example_without_comments_little_endian.ply'
    _test_file_with_invalid_comments_name = 'example_with_invalid_comments_little_endian.ply'



"""Feature extractor module."""
import copy
import itertools
import sys
import time

import numpy as np

from laserchicken import utils
from laserchicken.keys import point, provenance
from laserchicken.feature_extractor.base_feature_extractor import FeatureExtractor
from laserchicken.feature_extractor.feature_map import create_default_feature_map, _create_name_extractor_pairs

FEATURES = create_default_feature_map()


def list_feature_names():
    return FEATURES  # [feature_name for feature_name in FEATURES]


def register_new_feature_extractor(extractor: FeatureExtractor):
    for name, extractor in _create_name_extractor_pairs([extractor]):
        FEATURES[name] = extractor


def compute_features(env_point_cloud, neighborhoods, target_point_cloud, feature_names, volume, verbose=True, **kwargs):
    """
    Compute features for each target and store result as point attributes in target point cloud.

    Example:
    >>> point_cloud = load('data1.ply')
    >>> target_point_cloud = load('data2.ply')
    >>> volume = build_volume('infinite cylinder', radius=4)
    >>> neighbors = compute_neighborhoods(point_cloud, target_point_cloud, volume)
    >>> neighborhoods = []
    >>> for x in neighbors:
    >>>   neighborhoods += x
    >>> compute_features(point_cloud, neighborhoods, target_point_cloud, ['eigenv_1', 'kurto_z'], volume)
    >>> eigenv_1 = target_point_cloud[point]['eigenv_1']['data']

    Results of the example above are stored in the target point cloud as extra point attributes.

    :param env_point_cloud: environment point cloud
    :param neighborhoods: list of neighborhoods which are themselves lists of indices referring to the environment
    :param target_point_cloud: point cloud of targets
    :param feature_names: list of features that are to be calculated
    :param volume: object describing the volume that contains the neighborhood points
    :param kwargs: keyword arguments for the individual feature extractors
    :param verbose: if true, output extra information
    :return: None, results are stored in attributes of the target point cloud
    """
    _verify_feature_names(feature_names)
    wanted_feature_names = feature_names + [existing_feature for existing_feature in target_point_cloud[point]]
    extended_features = _make_extended_feature_list(feature_names)

    for feature_name in extended_features:
        target_point_cloud[point][feature_name] = {"type": 'float64',
                                                   "data": np.zeros_like(target_point_cloud[point]['x']['data'],
                                                                         dtype=np.float64)}

    if provenance in env_point_cloud:
        utils.add_metadata(target_point_cloud, sys.modules[__name__],
                           {'env_point_cloud': {provenance: copy.copy(env_point_cloud[provenance])}})

    _add_features(extended_features, env_point_cloud, neighborhoods, target_point_cloud, volume, verbose, kwargs)

    _keep_only_wanted_features(target_point_cloud, wanted_feature_names)


def _add_features(extended_features, env_point_cloud, neighborhoods, target_point_cloud, volume, verbose, kwargs):
    chunk_size = 100000
    n_targets = _get_point_cloud_size(target_point_cloud)
    for chunk_no in range(_calculate_number_of_chunks(chunk_size, n_targets)):
        i_start = chunk_no * chunk_size
        i_end = min((chunk_no + 1) * chunk_size, n_targets)
        target_indices = np.arange(i_start, i_end)
        current_neighborhoods = list(itertools.islice(neighborhoods, i_end - i_start))

        features_to_do = list(extended_features)

        _compute_features_for_chunk(features_to_do, env_point_cloud, current_neighborhoods, target_point_cloud,
                                    target_indices, volume, verbose, kwargs)


def _get_point_cloud_size(target_point_cloud):
    return len(target_point_cloud[point]['x']['data'])


def _calculate_number_of_chunks(chunk_size, n_targets):
    return int(np.math.ceil(n_targets / chunk_size))


def _compute_features_for_chunk(features_to_do, env_point_cloud, current_neighborhoods, target_point_cloud,
                                target_indices, volume,
                                verbose, kwargs):
    while features_to_do:
        feature_name = features_to_do[0]
        extractor = FEATURES[feature_name]

        if verbose:
            sys.stdout.write('Extracting feature(s) "{}"'.format(extractor.provides()))
            start = time.time()

        for key_word in kwargs:
            setattr(extractor, key_word, kwargs[key_word])
        _add_features_from_single_extractor(extractor, env_point_cloud, current_neighborhoods, target_point_cloud,
                                            target_indices, volume)
        utils.add_metadata(target_point_cloud, type(extractor).__module__, extractor.get_params())

        if verbose:
            elapsed = time.time() - start
            sys.stdout.write('Extracting feature(s) "{}" took {:.2f} seconds\n'.format(extractor.provides(), elapsed))

        for provided_feature in extractor.provides():
            if provided_feature in features_to_do:
                features_to_do.remove(provided_feature)


def _add_features_from_single_extractor(extractor, env_point_cloud, current_neighborhoods, target_point_cloud,
                                        target_indices, volume):
    provided_features = extractor.provides()
    n_features = len(provided_features)
    point_values = extractor.extract(env_point_cloud, current_neighborhoods, target_point_cloud,
                                     target_indices, volume)

    n_targets = len(target_indices)
    feature_values = [np.empty(n_targets, dtype=np.float64) for _ in range(n_features)]
    if n_features > 1:
        for i in range(n_features):
            feature_values[i] = point_values[i]
    else:
        feature_values[0] = point_values
    for i in range(n_features):
        feature = provided_features[i]
        target_point_cloud[point][feature]['data'][target_indices] = feature_values[i]


def _keep_only_wanted_features(target_point_cloud, wanted_feature_names):
    redundant_features = [f for f in target_point_cloud[point] if f not in wanted_feature_names]
    if redundant_features:
        print('The following unrequested features were calculated as a side effect, but will not be returned:',
              redundant_features)
    for f in redundant_features:
        target_point_cloud[point].pop(f)


def _verify_feature_names(feature_names):
    unknown_features = [f for f in feature_names if f not in FEATURES]
    if any(unknown_features):
        raise ValueError('Unknown features selected: {}. Available feature are: {}'
                         .format(', '.join(unknown_features), ', '.join(FEATURES.keys())))


def _make_extended_feature_list(feature_names):
    feature_list = reversed(_make_extended_feature_list_helper(feature_names))
    return _remove_duplicates(feature_list)


def _remove_duplicates(feature_list):
    seen = set()
    return [f for f in feature_list if not (f in seen or seen.add(f))]


def _make_extended_feature_list_helper(feature_names):
    feature_list = list(feature_names)
    for feature_name in feature_names:
        extractor = FEATURES[feature_name]
        dependencies = extractor.requires()
        feature_list.extend(dependencies)
        feature_list.extend(_make_extended_feature_list_helper(dependencies))
        feature_list.extend(extractor.provides())
    return feature_list



from unittest import TestCase

from laserchicken.build_volume import VOLUMES, build_volume
from laserchicken.volume_specification import Cell, Cube, InfiniteCylinder, Sphere, Volume


_shapes = [Cell, Cube, InfiniteCylinder, Sphere]


class BuildVolumeTest(TestCase):
    def test_volume_not_in_volume_map(self):
        self.assertNotIn(Volume, VOLUMES.values())

    def test_elements_in_volume_map(self):
        for shape in _shapes:
            self.assertIn(shape.TYPE, VOLUMES.keys())
            self.assertIs(VOLUMES[shape.TYPE], shape)

    def test_missing_build_volume_arguments(self):
        with self.assertRaises(TypeError):
            vol = build_volume()

    def test_build_nonexistent_volume(self):
        with self.assertRaises(ValueError):
            vol = build_volume("test", radius=5)

    def test_wrong_keyword_argument(self):
        with self.assertRaises(TypeError):
            vol = build_volume("sphere", cell_length=5)
        with self.assertRaises(TypeError):
            vol = build_volume("cell", radius=5)

    def test_build_all_elements(self):
        param = 5
        for shape in _shapes:
            vol = build_volume(shape.TYPE, param)
            self.assertIsInstance(vol, shape)
            for attr, val in shape(param).__dict__.items():
                self.assertIn(attr, vol.__dict__)
                self.assertAlmostEqual(val, vol.__dict__[attr])



"""Pulse penetration ratio and density absolute mean calculations.

See https://github.com/eEcoLiDAR/eEcoLiDAR/issues/23.
"""

import numpy as np

from laserchicken.feature_extractor.base_feature_extractor import FeatureExtractor
from laserchicken.keys import point, normalized_height

# classification according to
# http://www.asprs.org/wp-content/uploads/2010/12/LAS_1-4_R6.pdf
GROUND_TAGS = [2]


def _is_ground(i, point_cloud):
    return point_cloud[point]['raw_classification']["data"][i] in GROUND_TAGS


class PulsePenetrationFeatureExtractor(FeatureExtractor):
    """Feature extractor for the point density."""

    @classmethod
    def requires(cls):
        """
        Get a list of names of the point attributes that are needed for this feature extraction.

        For simple features, this could be just x, y, and z. Other features can build on again
        other features to have been computed first.

        :return: List of feature names
        """
        return []

    @classmethod
    def provides(cls):
        """
        Get a list of names of the feature values.

        This will return as many names as the number feature values that will be returned.
        For instance, if a feature extractor returns the first 3 eigen values, this method
        should return 3 names, for instance 'eigen_value_1', 'eigen_value_2' and 'eigen_value_3'.

        :return: List of feature names
        """
        return ['pulse_penetration_ratio']

    def extract(self, point_cloud, neighborhoods, target_point_cloud, target_indices, volume_description):
        """
        Extract the feature value(s) of the point cloud at location of the target.

        :param point_cloud: environment (search space) point cloud
        :param neighborhoods: list of arrays of indices of points within the point_cloud argument
        :param target_point_cloud: point cloud that contains target point
        :param target_indices: list of indices of the target point in the target point cloud
        :param volume_description: volume object that describes the shape and size of the search volume
        :return: feature values
        """
        return [self._extract_one(point_cloud, neighborhood) for neighborhood in neighborhoods]

    def _extract_one(self, point_cloud, neighborhood):
        if 'raw_classification' not in point_cloud[point]:
            raise ValueError(
                'Missing raw_classification attribute which is necessary for calculating pulse_penetratio and '
                'density_absolute_mean features.')

        ground_indices = [i for i in neighborhood if _is_ground(i, point_cloud)]
        pulse_penetration_ratio = self._get_pulse_penetration_ratio(
            ground_indices, len(neighborhood))

        return pulse_penetration_ratio

    @staticmethod
    def _get_ground_indices(point_cloud, ground_tags):
        index_grd = []
        for ipt, c in enumerate(point_cloud):
            if c in ground_tags:
                index_grd.append(ipt)
        return index_grd

    @staticmethod
    def _get_pulse_penetration_ratio(ground_indices, n_total_points):
        n_total = max(n_total_points, 1)
        n_ground = len(ground_indices)
        return float(n_ground) / n_total

    def get_params(self):
        """
        Return a tuple of parameters involved in the current feature extractor object.

        Needed for provenance.
        """
        return ()



import unittest

import numpy as np

from laserchicken import keys
from laserchicken.keys import point, normalized_height
from laserchicken.normalize import normalize
from laserchicken.test_tools import create_point_cloud
from laserchicken.utils import get_attribute_value


class TestNormalize(unittest.TestCase):
    def test_normalize_empty_point_cloud(self):
        point_cloud = create_point_cloud([], [], [])
        normalized_point_cloud = normalize(point_cloud)
        self.assertTrue(normalized_height in normalized_point_cloud[point])

    def test_normalize_tiny_equal_point_cloud(self):
        point_cloud = create_point_cloud([0, 0, 0], [0, 0, 0], [0, 0, 0])
        normalized_point_cloud = normalize(point_cloud)
        normalized_values = get_attribute_value(normalized_point_cloud, range(3), normalized_height)
        np.testing.assert_allclose(normalized_values, np.array([0, 0, 0]), atol=1e-7)

    def test_normalize_tiny_unequal_point_cloud(self):
        point_cloud = create_point_cloud([0, 0, 0], [0, 0, 0], [1, 2, 3])
        normalized_point_cloud = normalize(point_cloud)
        normalized_values = get_attribute_value(normalized_point_cloud, range(3), normalized_height)
        np.testing.assert_allclose(normalized_values, np.array([0, 1, 2]), atol=1e-7)

    def test_normalize_tiny_unequal_point_cloud_multiple_cells(self):
        """Last of the 3 points is not in the neighborhood of the others."""
        point_cloud = create_point_cloud([0, 0, 5], [0, 0, 0], [1, 2, 3])
        normalized_point_cloud = normalize(point_cloud, cell_size=2)
        normalized_values = get_attribute_value(normalized_point_cloud, range(3), normalized_height)
        np.testing.assert_allclose(normalized_values, np.array([0, 1, 0]), atol=1e-7)

    def test_normalize_provenance_data_present(self):
        """Last of the 3 points is not in the neighborhood of the others."""
        point_cloud = create_point_cloud([0, 0, 5], [0, 0, 0], [1, 2, 3])
        point_cloud.pop(keys.provenance, None)  # Remove any provenance data
        normalized_point_cloud = normalize(point_cloud, cell_size=2)
        self.assertTrue(keys.provenance in normalized_point_cloud)



import os
import random
import unittest

import numpy as np

from laserchicken import compute_features, keys, load, utils
from laserchicken.compute_neighbors import compute_cylinder_neighborhood
from laserchicken.feature_extractor.entropy_feature_extractor import EntropyFeatureExtractor
from laserchicken.feature_extractor.feature_extraction import compute_features
from laserchicken.test_tools import create_point_cloud
from laserchicken.volume_specification import InfiniteCylinder


class TestExtractEntropy(unittest.TestCase):
    _test_file_name = 'AHN3.las'
    _test_data_source = 'testdata'
    point_cloud = None

    def test_entropy_positive_value(self):
        """Test computing of eigenvalues in cylinder."""
        target_point_cloud = self._find_neighbors_for_random_targets_and_compute_entropy()
        n_targets = len(target_point_cloud[keys.point]["x"]["data"])

        for i in range(n_targets):
            entropy_z = utils.get_attribute_value(target_point_cloud, i, "entropy_z")
            self.assertTrue(entropy_z >= 0)

    def test_entropy_parameters_in_log(self):
        """Test if parameters were written to log."""
        expected_parameters = [0.1, None, None]
        target_point_cloud = self._find_neighbors_for_random_targets_and_compute_entropy()
        self.assertEqual(expected_parameters, target_point_cloud[keys.provenance][-1]["parameters"])

    def test_entropy_module_name_in_log(self):
        """Test if parameters were written to log."""
        target_point_cloud = self._find_neighbors_for_random_targets_and_compute_entropy()
        desired_module_name = 'laserchicken.feature_extractor.entropy_feature_extractor'
        self.assertEqual(desired_module_name, target_point_cloud[keys.provenance][-1]["module"])

    def _find_neighbors_for_random_targets_and_compute_entropy(self):
        num_all_pc_points = len(self.point_cloud[keys.point]["x"]["data"])
        rand_indices = [random.randint(0, num_all_pc_points) for _ in range(20)]
        target_point_cloud = utils.copy_point_cloud(self.point_cloud, rand_indices)
        radius = 25
        neighborhoods = list(compute_cylinder_neighborhood(self.point_cloud, target_point_cloud, radius))
        compute_features(self.point_cloud, neighborhoods, target_point_cloud, ["entropy_z"],
                         InfiniteCylinder(5), layer_thickness=0.1)
        return target_point_cloud

    def test_default_provides_correct(self):
        feature_names = EntropyFeatureExtractor().provides()
        self.assertIn('entropy_z', feature_names)

    def setUp(self):
        self.point_cloud = load(os.path.join(self._test_data_source, self._test_file_name))
        random.seed(102938482634)


class TestExtractNormalizedEntropy(unittest.TestCase):
    def test_use_norm_z(self):
        x = y = np.array([0, 0, 0])
        z = np.array([2, 2, 2])
        normalized_z = np.array([3, 4, 5])
        point_cloud = create_point_cloud(x, y, z, normalized_z=normalized_z)
        neighborhoods = [[0, 1, 2]]

        entropy = self.extractor.extract(point_cloud, neighborhoods, None, None, None)[0]

        self.assertNotAlmostEqual(entropy, 0)

    def test_default_provides_correct(self):
        feature_names = self.extractor.provides()
        self.assertIn('entropy_normalized_height', feature_names)

    def setUp(self):
        self.extractor = EntropyFeatureExtractor(data_key=keys.normalized_height)



import unittest

import numpy as np

from laserchicken import load, keys
from laserchicken.feature_extractor.range_feature_extractor import RangeFeatureExtractor
from laserchicken.test_tools import create_point_cloud


class TestRangeZFeatureExtractor(unittest.TestCase):
    def test_height_stats(self):
        pc_in = load("testdata/AHN2.las")
        neighborhood = [89664, 23893, 30638, 128795, 62052, 174453, 29129, 17127, 128215, 29667, 116156, 119157, 98591,
                        7018,
                        61494, 65194, 117931, 62971, 10474, 90322]
        max_z, min_z, range_z = RangeFeatureExtractor().extract(pc_in, [neighborhood], None, None, None)[:, 0]
        np.testing.assert_allclose(range_z, 5.5)
        np.testing.assert_allclose(max_z, 5.979999973773956)
        np.testing.assert_allclose(min_z, 0.47999997377395631)

    def test_height_stats_without_neighbors(self):
        pc_in = load("testdata/AHN2.las")
        neighborhood = []
        max_z, min_z, range_z = RangeFeatureExtractor().extract(pc_in, [neighborhood], pc_in, None, None)[:, 0]
        assert np.isnan(range_z)
        assert np.isnan(max_z)
        assert np.isnan(min_z)

    def test_default_provides_correct(self):
        feature_names = RangeFeatureExtractor().provides()
        self.assertIn('min_z', feature_names)
        self.assertIn('max_z', feature_names)
        self.assertIn('range_z', feature_names)


class TestRangeNormZFeatureExtractor(unittest.TestCase):
    def test_use_norm(self):
        x = y = np.array([0, 0, 0])
        z = np.array([2, 2, 2])
        normalized_z = np.array([3, 4, 5])
        point_cloud = create_point_cloud(x, y, z, normalized_z=normalized_z)
        neighborhood = [[0, 1, 2]]

        extractor = RangeFeatureExtractor(data_key=keys.normalized_height)
        _max, _min, _range = extractor.extract(point_cloud, neighborhood, None, None, None)

        self.assertAlmostEqual(_max, 5)
        self.assertAlmostEqual(_min, 3)
        self.assertAlmostEqual(_range, 2)

    def test_normalized_z_provides_correct(self):
        feature_names = RangeFeatureExtractor(data_key=keys.normalized_height).provides()
        self.assertIn('min_normalized_height', feature_names)
        self.assertIn('max_normalized_height', feature_names)
        self.assertIn('range_normalized_height', feature_names)



import os
import unittest

from laserchicken import keys,kd_tree,load


class TestKDTree(unittest.TestCase):

    _test_file_name = 'AHN3.las'
    _test_data_source = 'testdata'
    pointcloud = None

    def test_build_kd_tree(self):
        """ Should build kd tree without exception """
        kd_tree._build_kdtree(self.pointcloud)

    def test_kd_tree_cache(self):
        """ Tests the caching mechanism """
        firsttree = kd_tree.get_kdtree_for_pc(self.pointcloud)
        secondtree = kd_tree.get_kdtree_for_pc(self.pointcloud)
        self.assertEqual(firsttree,secondtree)

    def test_sphere_neighb_kd_tree(self):
        """ Tests whether sphere neighborhood gives good result """
        tree = kd_tree.get_kdtree_for_pc(self.pointcloud)
        rad = 10.
        point_index = 1000
        p = [self.pointcloud[keys.point]["x"]["data"][point_index],
             self.pointcloud[keys.point]["y"]["data"][point_index]]
        indices = tree.query_ball_point(p,rad)
        for i in indices:
            q = [self.pointcloud[keys.point]["x"]["data"][i],
                 self.pointcloud[keys.point]["y"]["data"][i]]
            self.assertTrue((p[0] - q[0])**2 + (p[1] - q[1])**2 < rad**2)



    def setUp(self):
        self.pointcloud = load(os.path.join(self._test_data_source,self._test_file_name))

    def tearDown(self):
        pass



from laserchicken.io import get_io_handler

def export(point_cloud, path, format=None, overwrite=False, *args, **kwargs):
    """
    Write point cloud data to a file.

    :param point_cloud:
    :param path:
    :param format: point cloud file format
    :param overwrite: if path exists, overwrite
    :param args: optional non-keyword arguments to be passed to the format-specific writer
    :param kwargs: optional keyword arguments to be passed to the format-specific writer
    """
    writer = get_io_handler(path, mode='w', format=format, overwrite=overwrite)
    writer.write(point_cloud, *args, **kwargs)


import itertools
import os
import unittest

import numpy as np

from laserchicken import load, keys
from laserchicken.feature_extractor.percentile_feature_extractor import PercentileFeatureExtractor
from laserchicken.test_tools import create_point_cloud


class TestPercentileZFeatureExtractorArtificialData(unittest.TestCase):
    """Test percentile feature extractor on artificial data."""

    def test_percentile_z(self):
        xyz = np.array([list(p) for p in list(itertools.product(np.linspace(0, 1, 11), repeat=3))])
        point_cloud = create_point_cloud(xyz[:, 0], xyz[:, 1], xyz[:, 2])
        expected = np.linspace(0.1, 1.0, 10)
        extractors = [PercentileFeatureExtractor(p) for p in range(10, 110, 10)]

        percentiles = [e.extract(point_cloud, [range(len(xyz))], None, None, None)[0] for e in extractors]

        np.testing.assert_allclose(percentiles, expected)

    def test_default_provides_correct(self):
        feature_names = PercentileFeatureExtractor(54).provides()
        self.assertIn('perc_54_z', feature_names)


class TestPercentileFeatureExtractorRealData(unittest.TestCase):
    """Test percentile feature extractor on real data."""

    def test_percentile(self):
        """Compute the percentile of a given selection."""
        _test_file_name = 'AHN3.las'
        _test_data_source = 'testdata'

        point_cloud = load(os.path.join(_test_data_source, _test_file_name))
        index = [
            89664, 23893, 30638, 128795, 62052, 174453, 29129, 17127, 128215, 29667, 116156, 119157, 98591, 7018, 61494,
            65194, 117931, 62971, 10474, 90322
        ]
        extractor = PercentileFeatureExtractor()

        extractor.extract(point_cloud, index, None, None, None)


class TestPercentileNormZFeatureExtractorArtificialData(unittest.TestCase):
    """Test percentile feature extractor on artificial data."""

    def test_percentile_norm_z(self):
        xyz = np.array([list(p) for p in list(itertools.product(np.linspace(0, 1, 11), repeat=3))])
        point_cloud = create_point_cloud(xyz[:, 0], xyz[:, 1], np.zeros_like(xyz[:, 2]), normalized_z=xyz[:, 2])
        expected = np.linspace(0.1, 1.0, 10)
        extractors = [PercentileFeatureExtractor(p, data_key=keys.normalized_height) for p in range(10, 110, 10)]

        percentiles = np.hstack([e.extract(point_cloud, [range(len(xyz))], None, None, None)[0] for e in extractors])

        np.testing.assert_allclose(percentiles, expected)

    def test_default_provides_correct(self):
        feature_names = PercentileFeatureExtractor(54, data_key=keys.normalized_height).provides()
        self.assertIn('perc_54_normalized_height', feature_names)



import numpy as np

from laserchicken.feature_extractor.base_feature_extractor import FeatureExtractor
from laserchicken.keys import point


class RangeFeatureExtractor(FeatureExtractor):
    """Calculates the max, min and range on the z axis."""

    DEFAULT_MAX = float('NaN')
    DEFAULT_MIN = float('NaN')

    def __init__(self, data_key='z'):
        self.data_key = data_key

    @classmethod
    def requires(cls):
        return []

    def provides(self):
        base_names = ['max_', 'min_', 'range_']
        return [base + str(self.data_key) for base in base_names]

    def extract(self, point_cloud, neighborhoods, target_point_cloud, target_indices, volume_description):
        return np.array([self._extract_one(point_cloud, neighborhood) for neighborhood in neighborhoods]).T

    def _extract_one(self, source_point_cloud, neighborhood):
        if neighborhood:
            source_data = source_point_cloud[point][self.data_key]['data'][neighborhood]
            max_z = np.max(source_data) if len(source_data) > 0 else self.DEFAULT_MAX
            min_z = np.min(source_data) if len(source_data) > 0 else self.DEFAULT_MIN
            range_z = max_z - min_z
        else:
            max_z = min_z = range_z = np.NaN
        return max_z, min_z, range_z



import numpy as np


_PLY_IMPLICIT_TO_SHORT_TYPES = {
    "uchar": "u1",
    "char": "i1",
    "ushort": "u2",
    "short": "i2",
    "uint": "u4",
    "int": "i4",
    "float": "f4",
    "double": "f8",
}

_SHORT_TO_SINGLE_CHAR_TYPES = {
    "u1": "B",
    "i1": "b",
    "u2": "H",
    "i2": "h",
    "u4": "I",
    "i4": "i",
    "u8": "q",
    "i8": "Q",
    "f4": "f",
    "f8": "d",
}


def select_valid_attributes(attributes_all, attributes_to_select):
    if attributes_to_select is None:
        raise ValueError('Invalid list of attributes provided.')
    else:
        if ((isinstance(attributes_to_select, str) and attributes_to_select == 'all')
                or 'all' in attributes_to_select):
            return attributes_all
        else:
            invalid_attributes = [el for el in attributes_to_select if el not in attributes_all]
            if not invalid_attributes:
                return ['x', 'y', 'z'] \
                       + [el for el in attributes_to_select if el in attributes_all and el not in 'xyz']
            else:
                raise ValueError('Invalid attributes provided: {}'.format(', '.join(invalid_attributes)))


def convert_to_short_type(type, use_ply_implicit=False):
    """
    Convert python type to short format

    Example:
    >>>> convert_to_short_type('float64')
    "f8"
    >>>> convert_to_short_type('float')
    "f8"
    >>>> convert_to_short_type('float', use_ply_implicit=True)
    "f4"

    :param type: python type string
    :param use_ply_implicit: if true, use implicit PLY types (e.g. float -> f4)
    :return: short type string
    """
    type_to_convert = type
    if (use_ply_implicit and
            type_to_convert in _PLY_IMPLICIT_TO_SHORT_TYPES.keys()):
        type_to_convert = _PLY_IMPLICIT_TO_SHORT_TYPES[type_to_convert]
    dtype = np.dtype(type_to_convert)
    return "".join([el for el in dtype.str
                    if el not in ["<", ">", "|"]])


def convert_to_single_character_type(type, use_ply_implicit=False):
    """
    Convert python type to single character type

    Example:
    >>>> convert_to_single_character_type('float64')
    "d"
    >>>> convert_to_single_character_type('f8')
    "d"
    >>>> convert_to_single_character_type('float')
    "d"
    >>>> convert_to_single_character_type('float', use_ply_implicit=True)
    "f"

    :param type: python type string
    :param use_ply_implicit: if true, use implicit PLY types (e.g. float -> f4)
    :return one character string
    """
    # convert string to np data type and remove byte order character
    type_short = convert_to_short_type(type, use_ply_implicit)
    if type_short not in _SHORT_TO_SINGLE_CHAR_TYPES.keys():
        raise ValueError("Type {} not known! Choose between: {}".format(type,
                                                                        ",".join(_SHORT_TO_SINGLE_CHAR_TYPES.keys())))
    return _SHORT_TO_SINGLE_CHAR_TYPES[type_short]



import numpy as np
import scipy.stats.stats as stat

from laserchicken.feature_extractor.base_feature_extractor import FeatureExtractor
from laserchicken.keys import point


class KurtosisFeatureExtractor(FeatureExtractor):
    """Calculates the variation on the z axis."""
    def __init__(self, data_key='z'):
        self.data_key = data_key

    @classmethod
    def requires(cls):
        return []

    def provides(self):
        return ['kurto_' + self.data_key]

    def extract(self, point_cloud, neighborhoods, target_point_cloud, target_indices, volume_description):
        return [self._extract_one(point_cloud, neighborhood) for neighborhood in neighborhoods]

    def _extract_one(self, sourcepc, neighborhood):
        if neighborhood:
            z = sourcepc[point][self.data_key]['data'][neighborhood]
            kurtosis_z = stat.kurtosis(z)
        else:
            kurtosis_z = np.NaN
        return kurtosis_z



import os
import random
import unittest

import numpy as np
from numpy.testing import assert_equal

from laserchicken import kd_tree, keys, load, utils
from laserchicken.compute_neighbors import compute_neighborhoods, compute_cylinder_neighborhood, \
    compute_sphere_neighborhood
from laserchicken.test_tools import create_point_cloud, create_points_in_xy_grid
from laserchicken.volume_specification import Sphere, InfiniteCylinder, Cell, Cube


class TestComputeNeighbors(unittest.TestCase):
    _test_file_name = 'AHN3.las'
    _test_data_source = 'testdata'
    point_cloud = None

    def test_compute_cylinder_neighbors(self):
        """Compute neighbors should only return points within the (xy) radius of the target."""
        target_point_cloud = self._get_random_targets()
        radius = 0.5
        neighborhoods = list(compute_cylinder_neighborhood(self.point_cloud, target_point_cloud, radius))
        self._assert_all_points_within_cylinder(neighborhoods, target_point_cloud, radius)

    def test_compute_sphere_neighbors(self):
        """Compute neighbors should only return points within the (xyz) radius of the target."""
        target_point_cloud = self._get_random_targets()
        radius = 0.5
        neighborhoods = list(compute_sphere_neighborhood(self.point_cloud, target_point_cloud, radius))
        self._assert_all_points_within_sphere(neighborhoods, target_point_cloud, radius)

    def test_compute_neighbors_sphereVolume(self):
        """Compute neighbors should detect sphere volume and find neighbors accordingly"""
        target_point_cloud = self._get_random_targets()
        sphere = Sphere(0.5)
        neighborhoods = compute_neighborhoods(self.point_cloud, target_point_cloud, sphere)
        self._assert_all_points_within_sphere(neighborhoods, target_point_cloud, sphere.radius)

    def test_compute_neighbors_cylinderVolume(self):
        """Compute neighbors should detect cylinder volume and find neighbors accordingly"""
        target_point_cloud = self._get_random_targets()
        cylinder = InfiniteCylinder(0.5)
        neighborhoods = compute_neighborhoods(self.point_cloud, target_point_cloud, cylinder)
        self._assert_all_points_within_cylinder(neighborhoods, target_point_cloud, cylinder.radius)

    def test_cell_no_points(self):
        point_cloud = create_emtpy_point_cloud()
        targets = create_point_cloud(np.zeros(1), np.zeros(1), np.zeros(1))
        neighborhoods = list(compute_neighborhoods(point_cloud, targets, Cell(1)))
        assert_equal(len(neighborhoods[0]), 0)

    def test_cell_grid(self):
        _, points = create_points_in_xy_grid(lambda x, y: np.random.rand())
        point_cloud = create_point_cloud(points[:, 0], points[:, 1], points[:, 2])
        targets = create_point_cloud(np.array([4.5]), np.array([4.5]), np.array([4.5]))  # Center of grid
        neighborhoods = list(compute_neighborhoods(point_cloud, targets, Cell(2)))
        assert_equal(len(neighborhoods[0]), 4)

    def test_cell_grid_sample_size(self):
        _, points = create_points_in_xy_grid(lambda x, y: np.random.rand())
        point_cloud = create_point_cloud(points[:, 0], points[:, 1], points[:, 2])
        targets = create_point_cloud(np.array([4.5]), np.array([4.5]), np.array([4.5]))  # Center of grid
        neighborhoods = list(compute_neighborhoods(point_cloud, targets, Cell(5), sample_size=3))
        assert_equal(len(neighborhoods[0]), 3)

    def test_cell_grid_larger_sample_size(self):
        _, points = create_points_in_xy_grid(lambda x, y: np.random.rand())
        point_cloud = create_point_cloud(points[:, 0], points[:, 1], points[:, 2])
        targets = create_point_cloud(np.array([4.5]), np.array([4.5]), np.array([4.5]))  # Center of grid
        neighborhoods = compute_neighborhoods(point_cloud, targets, Cell(5), sample_size=10000)  # Result 36 neighbors
        _ = next(neighborhoods)

    def test_cell_grid_origin(self):
        _, points = create_points_in_xy_grid(lambda x, y: np.random.rand())
        point_cloud = create_point_cloud(points[:, 0], points[:, 1], points[:, 2])
        targets = create_point_cloud(np.array([0]), np.array([0]), np.array([0]))  # Center of grid
        neighborhoods = list(compute_neighborhoods(point_cloud, targets, Cell(1.99)))
        assert_equal(len(neighborhoods[0]), 1)

    def test_cube_no_points(self):
        point_cloud = create_emtpy_point_cloud()
        targets = create_point_cloud(np.zeros(1), np.zeros(1), np.zeros(1))
        neighborhoods = list(compute_neighborhoods(point_cloud, targets, Cube(1)))
        assert_equal(len(neighborhoods[0]), 0)

    def test_cube_grid(self):
        _, points = create_points_in_xy_grid(lambda x, y: 10 * (x % 2))
        point_cloud = create_point_cloud(points[:, 0], points[:, 1], points[:, 2])
        targets = create_point_cloud(np.array([4.5]), np.array([4.5]), np.array([0]))  # Center of grid
        neighborhoods = list(compute_neighborhoods(point_cloud, targets, Cube(2)))
        assert_equal(len(neighborhoods[0]), 2)

    def test_target_number_matches_neighborhood_number(self):
        _, points = create_points_in_xy_grid(lambda x, y: 10 * (x % 2))
        environment_point_cloud = create_point_cloud(points[:, 0], points[:, 1], points[:, 2])
        assert_target_number_matches_neighborhood_number(environment_point_cloud)

    def test_target_number_matches_neighborhood_number_with_empty_point_cloud(self):
        environment_point_cloud = create_emtpy_point_cloud()
        assert_target_number_matches_neighborhood_number(environment_point_cloud)

    def setUp(self):
        self.point_cloud = load(os.path.join(
            self._test_data_source, self._test_file_name))
        random.seed(102938482634)
        kd_tree.initialize_cache()

    def tearDown(self):
        pass

    def _get_random_targets(self):
        num_all_pc_points = len(self.point_cloud[keys.point]["x"]["data"])
        rand_indices = [random.randint(0, num_all_pc_points) for _ in range(20)]
        return utils.copy_point_cloud(self.point_cloud, rand_indices)

    def _assert_all_points_within_cylinder(self, index_sets, target_point_cloud, radius):
        point_clouds = [utils.copy_point_cloud(self.point_cloud, indices) for indices in index_sets]
        n_targets = len(target_point_cloud[keys.point]["x"]["data"])
        assert_equal(n_targets, len(point_clouds))
        for i in range(n_targets):
            target_x = target_point_cloud[keys.point]["x"]["data"][i]
            target_y = target_point_cloud[keys.point]["y"]["data"][i]
            for j in range(len(point_clouds[i][keys.point]["x"]["data"])):
                neighbor_x = point_clouds[i][keys.point]["x"]["data"][j]
                neighbor_y = point_clouds[i][keys.point]["y"]["data"][j]
                dist = np.sqrt((neighbor_x - target_x) ** 2 +
                               (neighbor_y - target_y) ** 2)
                self.assertTrue(dist <= radius)

    def _assert_all_points_within_sphere(self, index_sets, target_point_cloud, radius):
        point_clouds = [utils.copy_point_cloud(
            self.point_cloud, indices) for indices in index_sets]
        n_targets = len(target_point_cloud[keys.point]["x"]["data"])
        assert_equal(n_targets, len(point_clouds))
        for i in range(n_targets):
            target_x, target_y, target_z = utils.get_point(
                target_point_cloud, i)
            for j in range(len(point_clouds[i][keys.point]["x"]["data"])):
                neighbor_x, neighbor_y, neighbor_z = utils.get_point(
                    point_clouds[i], j)
                dist = np.sqrt(
                    (neighbor_x - target_x) ** 2 + (neighbor_y - target_y) ** 2 + (neighbor_z - target_z) ** 2)
                self.assertTrue(dist <= radius)


def create_emtpy_point_cloud():
    empty = np.zeros(0)
    point_cloud = create_point_cloud(empty, empty, empty)
    return point_cloud


def assert_target_number_matches_neighborhood_number(environment_point_cloud):
    n_targets = 99
    targets = create_point_cloud(np.array(range(n_targets)), np.array(range(n_targets)), np.array(range(n_targets)))
    neighborhoods = list(compute_neighborhoods(environment_point_cloud, targets, Cube(2)))
    assert_equal(len(neighborhoods), n_targets)



""" Abstract IO handler """
import os


class IOHandler(object):
    """
    Abstract IO handler class

    Readers/writers should handle point-cloud data with the following structure:
    {'log': ['Processed by module load', 'Processed by module filter using parameters(x,y,z)'],
     'pointcloud':
        {'offset': {'type': 'double', 'data': 12.1}},
     'vertex':
        {'x': {'type': 'double', 'data': np.array([0.1, 0.2, 0.3])},
         'y': {'type': 'double', 'data': np.array([0.1, 0.2, 0.3])},
         'z': {'type': 'double', 'data': np.array([0.1, 0.2, 0.3])},
         'return': {'type': 'int', 'data': np.array([1, 1, 2])}}}
    """
    path = None

    def __init__(self, path, mode, overwrite=False):
        """
        Perform some checks on the path where we need to operate (read/write).

        :param path: path where IO needs to take place
        :param mode: 'r' for reading, 'w' for writing
        :param overwrite: if writing, overwrite path if it already exists
        """
        self.path = path
        if mode == 'r':
            if not os.path.exists(path):
                raise FileNotFoundError('{} not found.'.format(path))
        elif mode == 'w':
            path_directory = os.path.dirname(path)
            if path_directory and not os.path.exists(path_directory):
                raise FileNotFoundError('Output file path does not exist! --> {}'.format(path_directory))
            if os.path.exists(path):
                if not overwrite:
                    # Raise most specific subclass of FileExistsError (3.6) and IOError (2.7).
                    raise FileExistsError('Output file already exists! --> {}'.format(path))
                else:
                    os.remove(path)

    def read(self):
        """
        Read the point cloud from disk

        :return point_cloud:
        """
        raise NotImplementedError(
            "Class %s doesn't implement read()" % self.__class__.__name__)

    def write(self, point_cloud):
        """
        Write the point cloud to disk

        :param point_cloud:
        """
        raise NotImplementedError(
            "Class %s doesn't implement write()" % self.__class__.__name__)



import unittest

import numpy as np

from laserchicken import keys
from laserchicken.feature_extractor.density_absolute_mean_feature_extractor import DensityAbsoluteMeanFeatureExtractor
from laserchicken.keys import point
from laserchicken.test_tools import create_point_cloud


class TestDensityAbsoluteMeanZFeatureExtractorArtificialData(unittest.TestCase):
    def test_simle_case_correct(self):
        """Check that one out of 4 points above mean of only vegetation points yields a value of 25"""
        ground = 2  # Ground tag
        veg = 4  # Medium vegetation tag
        x = y = z = np.array([10, 10, 10, 1, 1, 1, 2])
        point_cloud = create_point_cloud(x, y, z)
        point_cloud[point]['raw_classification'] = {'data': np.array([ground, ground, ground, veg, veg, veg, veg]),
                                                    'type': 'double'}
        neighborhood = list(range(len(x)))

        density_absolute_mean = self.extractor.extract(point_cloud, [neighborhood], None, None, None)[0]

        self.assertAlmostEqual(density_absolute_mean, 25)

    def test_default_provides_correct(self):
        feature_names = self.extractor.provides()
        self.assertIn('density_absolute_mean_z', feature_names)

    def setUp(self):
        self.extractor = DensityAbsoluteMeanFeatureExtractor()


class TestDensityAbsoluteMeanNormZFeatureExtractorArtificialData(unittest.TestCase):
    def test_simple_case_correct(self):
        """Check that one out of 4 points above mean of only vegetation points yields a value of 25"""
        ground = 2  # Ground tag
        veg = 4  # Medium vegetation tag
        x = y = z = np.array([10, 10, 10, 1, 1, 1, 2])
        point_cloud = create_point_cloud(x, y, np.zeros_like(z), normalized_z=z)
        point_cloud[point]['raw_classification'] = {'data': np.array([ground, ground, ground, veg, veg, veg, veg]),
                                                    'type': 'double'}
        neighborhood = list(range(len(x)))

        density_absolute_mean = self.extractor.extract(point_cloud, [neighborhood], None, None, None)[0]

        self.assertAlmostEqual(density_absolute_mean, 25)

    def test_default_provides_correct(self):
        feature_names = self.extractor.provides()
        self.assertIn('density_absolute_mean_normalized_height', feature_names)

    def setUp(self):
        self.extractor = DensityAbsoluteMeanFeatureExtractor(data_key=keys.normalized_height)
